import pytest
from pytest_mock import MockFixture

from cg.apps.lims import LimsAPI
from cg.meta.upload.fohm.fohm import FOHMUploadAPI
from cg.models.cg_config import CGConfig
from cg.models.fohm.reports import FohmComplementaryReport, FohmPangolinReport
from cg.store.store import Store
from tests.store_helpers import StoreHelpers


@pytest.fixture
def fohm_complementary_report_raw() -> dict[str, str]:
    """Return a raw FOHM complementary report."""
    return {
        "provnummer": "a_sample_number",
        "urvalskriterium": "a_selection_criteria",
        "GISAID_accession": "an_accession",
    }


@pytest.fixture
def fohm_complementary_report(
    fohm_complementary_report_raw: list[dict],
) -> FohmComplementaryReport:
    """Return FohmComplementaryReport."""
    return FohmComplementaryReport.model_validate(fohm_complementary_report_raw)


@pytest.fixture
def fohm_complementary_reports(
    fohm_complementary_report: FohmComplementaryReport,
) -> list[FohmComplementaryReport]:
    """Return FOHM complementary reports."""
    report_1 = FohmComplementaryReport.model_validate(
        {"provnummer": "1CS", "urvalskriterium": "criteria", "GISAID_accession": "an_accession"}
    )
    complementary_report = FohmComplementaryReport.model_validate(
        {
            "urvalskriterium": "criteria",
            "GISAID_accession": "an_accession",
            "provnummer": "44CS000000",
        }
    )
    return [fohm_complementary_report, complementary_report, report_1]


@pytest.fixture
def fohm_pangolin_report_raw() -> dict[str, str]:
    """Return a raw FOHM Pangolin report."""
    return {
        "taxon": "a_taxon",
        "lineage": "a_lineage",
        "conflict": "a_conflict",
        "ambiguity_score": "an_ambiguity_score",
        "scorpio_call": "a_scorpio_call",
        "scorpio_support": "a_scorpio_support",
        "scorpio_conflict": "a_scorpio_conflict",
        "scorpio_notes": "a_scorpio_note",
        "version": "a_version",
        "pangolin_version": "a_pangolin_version",
        "scorpio_version": "a_wcorpio_version",
        "constellation_version": "a_constellation_vversion",
        "is_designated": "a_designation",
        "qc_status": "a_qc_status",
        "qc_notes": "a_qc_note",
        "note": "a_note",
    }


@pytest.fixture
def fohm_pangolin_report(
    fohm_pangolin_report_raw: list[dict],
) -> FohmPangolinReport:
    """Return FohmPangolinReport."""
    return FohmPangolinReport.model_validate(fohm_pangolin_report_raw)


@pytest.fixture
def fohm_pangolin_reports(
    fohm_pangolin_report: FohmPangolinReport,
) -> list[FohmPangolinReport]:
    """Return FOHM pangolin reports."""
    pangolin_report = FohmPangolinReport.model_validate(
        {
            "taxon": "44CS000001",
            "lineage": "a_lineage",
            "conflict": "a_conflict",
            "ambiguity_score": "an_ambiguity_score",
            "scorpio_call": "a_scorpio_call",
            "scorpio_support": "a_scorpio_support",
            "scorpio_conflict": "a_scorpio_conflict",
            "scorpio_notes": "a_scorpio_note",
            "version": "a_version",
            "pangolin_version": "a_pangolin_version",
            "scorpio_version": "a_wcorpio_version",
            "constellation_version": "a_constellation_vversion",
            "is_designated": "a_designation",
            "qc_status": "a_qc_status",
            "qc_notes": "a_qc_note",
            "note": "a_note",
        }
    )
    return [fohm_pangolin_report, pangolin_report]


@pytest.fixture
def fohm_upload_api(
    cg_context: CGConfig, mocker: MockFixture, helpers: StoreHelpers
) -> FOHMUploadAPI:
    """FOHM upload API fixture."""
    fohm_upload_api = FOHMUploadAPI(cg_context)

    # Mock getting Sample object from StatusDB
    mocker.patch.object(
        Store,
        "get_sample_by_name",
        return_value=helpers.add_sample(fohm_upload_api.status_db, internal_id="a_sample_name"),
    )

    # Mock getting sample attribute from LIMS
    mocker.patch.object(LimsAPI, "get_sample_attribute", return_value="a_sample_attribute")

    return fohm_upload_api



from datetime import datetime

import pytest

from cg.constants.devices import DeviceType
from cg.services.illumina.data_transfer.models import (
    IlluminaSampleSequencingMetricsDTO,
    IlluminaSequencingRunDTO,
)


@pytest.fixture(scope="session")
def empty_illumina_sequencing_dto() -> IlluminaSequencingRunDTO:
    """Return an empty IlluminaSequencingRunDTO."""
    return IlluminaSequencingRunDTO(
        type=DeviceType.ILLUMINA,
        sequencer_type=None,
        sequencer_name=None,
        data_availability=None,
        archived_at=None,
        has_backup=None,
        total_reads=None,
        total_undetermined_reads=None,
        percent_undetermined_reads=None,
        percent_q30=None,
        mean_quality_score=None,
        total_yield=None,
        yield_q30=None,
        cycles=None,
        demultiplexing_software=None,
        demultiplexing_software_version=None,
        sequencing_started_at=None,
        sequencing_completed_at=None,
        demultiplexing_started_at=None,
        demultiplexing_completed_at=None,
    )


@pytest.fixture
def mapped_metric(timestamp_now: datetime):
    return IlluminaSampleSequencingMetricsDTO(
        sample_id="sample",
        type=DeviceType.ILLUMINA,
        flow_cell_lane=1,
        total_reads_in_lane=100,
        base_passing_q30_percent=0.9,
        base_mean_quality_score=30,
        yield_=100,
        yield_q30=0.9,
        created_at=timestamp_now,
    )


@pytest.fixture
def undetermined_metric(timestamp_now: datetime):
    return IlluminaSampleSequencingMetricsDTO(
        sample_id="sample",
        flow_cell_lane=1,
        type=DeviceType.ILLUMINA,
        total_reads_in_lane=100,
        base_passing_q30_percent=0.8,
        base_mean_quality_score=20,
        yield_=100,
        yield_q30=0.8,
        created_at=timestamp_now,
    )


@pytest.fixture
def combined_metric(timestamp_now: datetime):
    return IlluminaSampleSequencingMetricsDTO(
        sample_id="sample",
        flow_cell_lane=1,
        type=DeviceType.ILLUMINA,
        total_reads_in_lane=200,
        base_passing_q30_percent=0.85,
        base_mean_quality_score=25,
        yield_=200,
        yield_q30=0.85,
        created_at=timestamp_now,
    )



from pathlib import Path

import pytest
from dateutil.parser import parse

from cg.constants import Workflow
from cg.meta.workflow.rnafusion import RnafusionAnalysisAPI
from cg.meta.workflow.taxprofiler import TaxprofilerAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.store.models import Analysis


@pytest.mark.parametrize(
    "before_date, should_be_cleaned", [("2024-12-31", True), ("2024-01-01", False)]
)
def test_clean_rnafusion_case_directories(
    before_date: str, should_be_cleaned: bool, rnafusion_context: CGConfig
):
    """Tests cleaning of case directories."""

    # GIVEN a file which was modified days_since_modification ago
    analysis_to_clean: Analysis = rnafusion_context.status_db.add_analysis(
        workflow=Workflow.RNAFUSION, started_at=parse("2024-06-06"), uploaded=parse("2024-06-06")
    )
    analysis_to_clean.case = rnafusion_context.status_db.get_cases()[0]
    rnafusion_context.status_db.session.add(analysis_to_clean)
    rnafusion_context.status_db.commit_to_store()

    # GIVEN an RNAFusion analysis API
    analysis_api: RnafusionAnalysisAPI = rnafusion_context.meta_apis["analysis_api"]

    case_path = analysis_api.get_case_path(analysis_to_clean.case.internal_id)
    case_path.mkdir()
    file_path = Path(case_path, "test_file")
    file_path.touch()

    # WHEN cleaning the tmp_path/cases directory of old files
    analysis_api.clean_past_run_dirs(before_date=before_date, skip_confirmation=True)

    # THEN the file should be deleted if it was old.
    assert file_path.exists() != should_be_cleaned


@pytest.mark.parametrize(
    "before_date, should_be_cleaned", [("2024-12-31", True), ("2024-01-01", False)]
)
def test_clean_taxprofiler_case_directories(
    before_date: str, should_be_cleaned: bool, taxprofiler_context: CGConfig
):
    """Tests cleaning of case directories."""

    # GIVEN a file which was modified days_since_modification ago
    analysis_to_clean: Analysis = taxprofiler_context.status_db.add_analysis(
        workflow=Workflow.TAXPROFILER, started_at=parse("2024-06-06"), uploaded=parse("2024-06-06")
    )
    analysis_to_clean.case = taxprofiler_context.status_db.get_cases()[0]
    taxprofiler_context.status_db.session.add(analysis_to_clean)
    taxprofiler_context.status_db.commit_to_store()

    # GIVEN a Taxprofiler analysis API
    analysis_api: TaxprofilerAnalysisAPI = taxprofiler_context.meta_apis["analysis_api"]

    case_path = analysis_api.get_case_path(analysis_to_clean.case.internal_id)
    case_path.mkdir()
    file_path = Path(case_path, "test_file")
    file_path.touch()

    # WHEN cleaning the tmp_path/cases directory of old files
    analysis_api.clean_past_run_dirs(before_date=before_date, skip_confirmation=True)

    # THEN the file should be deleted if it was old.
    assert file_path.exists() != should_be_cleaned



import logging
from datetime import datetime

from cg.constants import DataDelivery, GenePanelMasterList, Priority, Workflow
from cg.constants.constants import CustomerId
from cg.constants.sequencing import SeqLibraryPrepCategory
from cg.exc import OrderError
from cg.models.orders.order import OrderIn
from cg.models.orders.sample_base import StatusEnum
from cg.services.orders.order_lims_service.order_lims_service import OrderLimsService
from cg.services.orders.submitters.order_submitter import StoreOrderService
from cg.store.models import ApplicationVersion, Case, CaseSample, Customer, Order, Sample
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class StoreFastqOrderService(StoreOrderService):
    """Storing service for FASTQ orders."""

    def __init__(self, status_db: Store, lims_service: OrderLimsService):
        self.status_db = status_db
        self.lims = lims_service

    def store_order(self, order: OrderIn) -> dict:
        """Submit a batch of samples for FASTQ delivery."""

        project_data, lims_map = self.lims.process_lims(lims_order=order, new_samples=order.samples)
        status_data = self.order_to_status(order)
        self._fill_in_sample_ids(samples=status_data["samples"], lims_map=lims_map)
        new_samples = self.store_items_in_status(
            customer_id=status_data["customer"],
            order=status_data["order"],
            ordered=project_data["date"],
            ticket_id=order.ticket,
            items=status_data["samples"],
        )
        return {"project": project_data, "records": new_samples}

    @staticmethod
    def order_to_status(order: OrderIn) -> dict:
        """Convert order input to status for fastq-only orders."""
        status_data = {
            "customer": order.customer,
            "order": order.name,
            "samples": [
                {
                    "application": sample.application,
                    "capture_kit": sample.capture_kit,
                    "comment": sample.comment,
                    "data_analysis": sample.data_analysis,
                    "data_delivery": sample.data_delivery,
                    "name": sample.name,
                    "priority": sample.priority,
                    "sex": sample.sex,
                    "subject_id": sample.subject_id,
                    "tumour": sample.tumour,
                    "volume": sample.volume,
                }
                for sample in order.samples
            ],
        }
        return status_data

    def create_maf_case(self, sample_obj: Sample, order: Order) -> None:
        """Add a MAF case to the Status database."""
        case: Case = self.status_db.add_case(
            data_analysis=Workflow(Workflow.MIP_DNA),
            data_delivery=DataDelivery(DataDelivery.NO_DELIVERY),
            name="_".join([sample_obj.name, "MAF"]),
            panels=[GenePanelMasterList.OMIM_AUTO],
            priority=Priority.research,
            ticket=sample_obj.original_ticket,
        )
        case.customer = self.status_db.get_customer_by_internal_id(
            customer_internal_id=CustomerId.CG_INTERNAL_CUSTOMER
        )
        relationship: CaseSample = self.status_db.relate_sample(
            case=case, sample=sample_obj, status=StatusEnum.unknown
        )
        order.cases.append(case)
        self.status_db.session.add_all([case, relationship])

    def store_items_in_status(
        self, customer_id: str, order: str, ordered: datetime, ticket_id: str, items: list[dict]
    ) -> list[Sample]:
        """Store fastq samples in the status database including family connection and delivery"""
        customer: Customer = self.status_db.get_customer_by_internal_id(
            customer_internal_id=customer_id
        )
        if not customer:
            raise OrderError(f"Unknown customer: {customer_id}")
        new_samples = []
        case: Case = self.status_db.get_case_by_name_and_customer(
            customer=customer, case_name=ticket_id
        )
        submitted_case: dict = items[0]
        status_db_order = Order(
            customer=customer,
            order_date=datetime.now(),
            ticket_id=int(ticket_id),
        )
        with self.status_db.session.no_autoflush:
            for sample in items:
                new_sample = self.status_db.add_sample(
                    name=sample["name"],
                    sex=sample["sex"] or "unknown",
                    comment=sample["comment"],
                    internal_id=sample.get("internal_id"),
                    order=order,
                    ordered=ordered,
                    original_ticket=ticket_id,
                    priority=sample["priority"],
                    tumour=sample["tumour"],
                    capture_kit=sample["capture_kit"],
                    subject_id=sample["subject_id"],
                )
                new_sample.customer: Customer = customer
                application_tag: str = sample["application"]
                application_version: ApplicationVersion = (
                    self.status_db.get_current_application_version_by_tag(tag=application_tag)
                )
                if application_version is None:
                    raise OrderError(f"Invalid application: {sample['application']}")
                new_sample.application_version: ApplicationVersion = application_version
                new_samples.append(new_sample)
                if not case:
                    case = self.status_db.add_case(
                        data_analysis=Workflow(submitted_case["data_analysis"]),
                        data_delivery=DataDelivery(submitted_case["data_delivery"]),
                        name=ticket_id,
                        panels=None,
                        priority=submitted_case["priority"],
                        ticket=ticket_id,
                    )
                if (
                    not new_sample.is_tumour
                    and new_sample.prep_category == SeqLibraryPrepCategory.WHOLE_GENOME_SEQUENCING
                ):
                    self.create_maf_case(sample_obj=new_sample, order=status_db_order)
                case.customer = customer
                new_relationship = self.status_db.relate_sample(
                    case=case, sample=new_sample, status=StatusEnum.unknown
                )
                self.status_db.session.add_all([case, new_relationship])
        status_db_order.cases.append(case)
        self.status_db.session.add(status_db_order)
        self.status_db.session.add_all(new_samples)
        self.status_db.session.commit()
        return new_samples



"""Fixtures for the upload Scout API tests."""

import logging
from datetime import datetime
from pathlib import Path
from typing import Generator

import pytest
from housekeeper.store.models import Version

from cg.constants import DataDelivery, Workflow
from cg.constants.constants import FileFormat, GenomeVersion
from cg.constants.housekeeper_tags import HK_DELIVERY_REPORT_TAG, AnalysisTag, NFAnalysisTags
from cg.constants.pedigree import Pedigree
from cg.constants.scout import UploadTrack
from cg.constants.sequencing import SeqLibraryPrepCategory
from cg.constants.subject import PhenotypeStatus
from cg.io.controller import ReadFile
from cg.meta.upload.scout.balsamic_config_builder import BalsamicConfigBuilder
from cg.meta.upload.scout.mip_config_builder import MipConfigBuilder
from cg.meta.upload.scout.raredisease_config_builder import RarediseaseConfigBuilder
from cg.meta.upload.scout.uploadscoutapi import UploadScoutAPI
from cg.meta.workflow.raredisease import RarediseaseAnalysisAPI
from cg.models.analysis import NextflowAnalysis
from cg.models.cg_config import CGConfig
from cg.models.scout.scout_load_config import MipLoadConfig
from cg.store.models import Analysis, Case, Sample
from cg.store.store import Store
from tests.mocks.balsamic_analysis_mock import MockBalsamicAnalysis

# Mocks
from tests.mocks.hk_mock import MockHousekeeperAPI
from tests.mocks.limsmock import MockLimsAPI
from tests.mocks.madeline import MockMadelineAPI
from tests.mocks.mip_analysis_mock import MockMipAnalysis
from tests.mocks.scout import MockScoutAPI
from tests.store_helpers import StoreHelpers

LOG = logging.getLogger(__name__)


@pytest.fixture
def rna_case_id() -> str:
    """Return a rna case id."""
    return "affirmativecat"


@pytest.fixture
def dna_case_id(case_id: str) -> str:
    """Return a DNA case id."""
    return case_id


@pytest.fixture
def rna_sample_son_id() -> str:
    """Return a son RNA sample id."""
    return "rna_son"


@pytest.fixture
def rna_sample_daughter_id() -> str:
    """Return a daughter RNA sample id."""
    return "rna_daughter"


@pytest.fixture
def rna_sample_mother_id() -> str:
    """Return a mother RNA sample id."""
    return "rna_mother"


@pytest.fixture
def rna_sample_father_id() -> str:
    """Return a father RNA sample id."""
    return "rna_father"


@pytest.fixture
def dna_sample_son_id() -> str:
    """Return a son DNA sample id."""
    return "dna_son"


@pytest.fixture
def dna_sample_daughter_id() -> str:
    """Return a daughter DNA sample id."""
    return "dna_daughter"


@pytest.fixture
def dna_sample_mother_id() -> str:
    """Return a mother DNA sample id."""
    return "dna_mother"


@pytest.fixture
def dna_sample_father_id() -> str:
    """Return a father DNA sample id."""
    return "dna_father"


@pytest.fixture
def another_rna_sample_id() -> str:
    """Return another RNA sample id."""
    return "another_rna_sample_id"


@pytest.fixture
def rna_store(
    base_store: Store,
    helpers: StoreHelpers,
    rna_case_id: str,
    dna_case_id: str,
    rna_sample_son_id: str,
    rna_sample_daughter_id: str,
    rna_sample_father_id: str,
    rna_sample_mother_id: str,
    dna_sample_son_id: str,
    dna_sample_daughter_id: str,
    dna_sample_father_id: str,
    dna_sample_mother_id: str,
) -> Store:
    """Populate store with an RNA case that is connected to a DNA case via subject ID."""

    store: Store = base_store

    # an existing RNA case with related sample
    rna_case = helpers.ensure_case(
        store=store,
        case_name="rna_case",
        customer=helpers.ensure_customer(store=store),
        data_analysis=Workflow.MIP_RNA,
        data_delivery=DataDelivery.SCOUT,
    )
    rna_case.internal_id = rna_case_id

    rna_sample_son = helpers.add_sample(
        store=store,
        application_type=SeqLibraryPrepCategory.WHOLE_TRANSCRIPTOME_SEQUENCING,
        name=rna_sample_son_id,
        subject_id="son",
    )
    rna_sample_daughter = helpers.add_sample(
        store=store,
        application_type=SeqLibraryPrepCategory.WHOLE_TRANSCRIPTOME_SEQUENCING,
        name=rna_sample_daughter_id,
        subject_id="daughter",
    )
    rna_sample_mother = helpers.add_sample(
        store=store,
        application_type=SeqLibraryPrepCategory.WHOLE_TRANSCRIPTOME_SEQUENCING,
        name=rna_sample_mother_id,
        subject_id=Pedigree.MOTHER,
    )
    rna_sample_father = helpers.add_sample(
        store=store,
        application_type=SeqLibraryPrepCategory.WHOLE_TRANSCRIPTOME_SEQUENCING,
        name=rna_sample_father_id,
        subject_id=Pedigree.FATHER,
    )
    helpers.add_relationship(
        store=store,
        sample=rna_sample_son,
        case=rna_case,
        mother=rna_sample_mother,
        father=rna_sample_father,
        status=PhenotypeStatus.AFFECTED,
    )
    helpers.add_relationship(
        store=store,
        sample=rna_sample_daughter,
        case=rna_case,
        mother=rna_sample_mother,
        father=rna_sample_father,
        status=PhenotypeStatus.UNAFFECTED,
    )
    helpers.add_relationship(
        store=store, sample=rna_sample_mother, case=rna_case, status=PhenotypeStatus.UNAFFECTED
    )
    helpers.add_relationship(
        store=store, sample=rna_sample_father, case=rna_case, status=PhenotypeStatus.AFFECTED
    )

    for link in rna_case.links:
        link.sample.internal_id = link.sample.name

    # an existing DNA case with related sample
    dna_case = helpers.ensure_case(
        store=store,
        case_name="dna_case",
        customer=helpers.ensure_customer(store=store),
        data_analysis=Workflow.MIP_DNA,
        data_delivery=DataDelivery.SCOUT,
    )
    dna_case.internal_id = dna_case_id

    dna_sample_son = helpers.add_sample(
        store=store,
        application_tag=SeqLibraryPrepCategory.WHOLE_GENOME_SEQUENCING,
        application_type=SeqLibraryPrepCategory.WHOLE_GENOME_SEQUENCING,
        name=dna_sample_son_id,
        subject_id="son",
    )
    dna_sample_daughter = helpers.add_sample(
        store=store,
        application_tag=SeqLibraryPrepCategory.WHOLE_GENOME_SEQUENCING,
        application_type=SeqLibraryPrepCategory.WHOLE_GENOME_SEQUENCING,
        name=dna_sample_daughter_id,
        subject_id="daughter",
    )
    dna_sample_mother = helpers.add_sample(
        store=store,
        application_tag=SeqLibraryPrepCategory.WHOLE_GENOME_SEQUENCING,
        application_type=SeqLibraryPrepCategory.WHOLE_GENOME_SEQUENCING,
        name=dna_sample_mother_id,
        subject_id=Pedigree.MOTHER,
    )
    dna_sample_father = helpers.add_sample(
        store=store,
        application_tag=SeqLibraryPrepCategory.WHOLE_GENOME_SEQUENCING,
        application_type=SeqLibraryPrepCategory.WHOLE_GENOME_SEQUENCING,
        name=dna_sample_father_id,
        subject_id=Pedigree.FATHER,
    )
    helpers.add_relationship(
        store=store,
        sample=dna_sample_son,
        case=dna_case,
        mother=dna_sample_mother,
        father=dna_sample_father,
        status=PhenotypeStatus.AFFECTED,
    )
    helpers.add_relationship(
        store=store,
        sample=dna_sample_daughter,
        case=dna_case,
        mother=dna_sample_mother,
        father=dna_sample_father,
        status=PhenotypeStatus.UNAFFECTED,
    )
    helpers.add_relationship(
        store=store, sample=dna_sample_mother, case=dna_case, status=PhenotypeStatus.UNAFFECTED
    )
    helpers.add_relationship(
        store=store, sample=dna_sample_father, case=dna_case, status=PhenotypeStatus.AFFECTED
    )

    for link in dna_case.links:
        link.sample.internal_id = link.sample.name

    helpers.add_analysis(store=store, case=dna_case, uploaded_at=datetime.now())

    store.session.commit()
    return store


@pytest.fixture(name="lims_family")
def lims_family(fixtures_dir) -> dict:
    """Returns a LIMS-like case of samples."""
    return ReadFile.get_content_from_file(
        file_format=FileFormat.JSON,
        file_path=Path(fixtures_dir, "report", "lims_family." + FileFormat.JSON),
    )


@pytest.fixture(name="lims_samples")
def lims_samples(lims_family: dict) -> list[dict]:
    """Returns the samples of a LIMS case."""
    return lims_family["samples"]


@pytest.fixture(scope="function")
def mip_dna_analysis_hk_bundle_data(
    case_id: str,
    timestamp: datetime,
    mip_dna_analysis_dir: Path,
    sample_id: str,
    snv_vcf_file: str,
    sv_vcf_file: str,
    snv_research_vcf_file: str,
    sv_research_vcf_file: str,
    delivery_report_html: Path,
) -> dict:
    """Return MIP DNA bundle data for Housekeeper."""
    return {
        "name": case_id,
        "created": timestamp,
        "expires": timestamp,
        "files": [
            {
                "path": Path(mip_dna_analysis_dir, snv_vcf_file).as_posix(),
                "archive": False,
                "tags": ["vcf-snv-clinical"],
            },
            {
                "path": Path(mip_dna_analysis_dir, sv_vcf_file).as_posix(),
                "archive": False,
                "tags": ["vcf-sv-clinical"],
            },
            {
                "path": Path(mip_dna_analysis_dir, snv_research_vcf_file).as_posix(),
                "archive": False,
                "tags": ["vcf-snv-research"],
            },
            {
                "path": Path(mip_dna_analysis_dir, sv_research_vcf_file).as_posix(),
                "archive": False,
                "tags": ["vcf-sv-research"],
            },
            {
                "path": Path(mip_dna_analysis_dir, "str.vcf").as_posix(),
                "archive": False,
                "tags": ["vcf-str"],
            },
            {
                "path": Path(mip_dna_analysis_dir, "smn.vcf").as_posix(),
                "archive": False,
                "tags": ["smn-calling"],
            },
            {
                "path": Path(mip_dna_analysis_dir, "adm1.cram").as_posix(),
                "archive": False,
                "tags": ["cram", sample_id],
            },
            {
                "path": Path(mip_dna_analysis_dir, "report.pdf").as_posix(),
                "archive": False,
                "tags": ["delivery-report"],
            },
            {
                "path": Path(mip_dna_analysis_dir, "adm1.mt.bam").as_posix(),
                "archive": False,
                "tags": ["bam-mt", sample_id],
            },
            {
                "path": Path(mip_dna_analysis_dir, "vcf2cytosure.txt").as_posix(),
                "archive": False,
                "tags": ["vcf2cytosure", sample_id],
            },
            {
                "path": Path(mip_dna_analysis_dir, "multiqc.html").as_posix(),
                "archive": False,
                "tags": ["multiqc-html", sample_id],
            },
        ],
    }


@pytest.fixture(scope="function")
def mip_rna_analysis_hk_bundle_data(
    rna_case_id: str,
    timestamp: datetime,
    mip_dna_analysis_dir: Path,
    rna_sample_son_id: str,
    rna_sample_daughter_id: str,
    rna_sample_mother_id: str,
    rna_sample_father_id: str,
) -> dict:
    """Return MIP RNA bundle data for Housekeeper."""

    files: list[dict] = [
        {
            "path": Path(mip_dna_analysis_dir, f"{rna_case_id}.cram").as_posix(),
            "archive": False,
            "tags": ["cram", rna_case_id],
        },
        {
            "path": Path(mip_dna_analysis_dir, f"{rna_case_id}_report.selected.pdf").as_posix(),
            "archive": False,
            "tags": ["fusion", "pdf", "clinical", rna_case_id],
        },
        {
            "path": Path(mip_dna_analysis_dir, f"{rna_case_id}_report.pdf").as_posix(),
            "archive": False,
            "tags": ["fusion", "pdf", "research", rna_case_id],
        },
    ]
    for sample_id in [
        rna_sample_son_id,
        rna_sample_daughter_id,
        rna_sample_mother_id,
        rna_sample_father_id,
    ]:
        files.extend(
            [
                {
                    "path": Path(
                        mip_dna_analysis_dir, f"{sample_id}_lanes_1_star_sorted_sj.bigWig"
                    ).as_posix(),
                    "archive": False,
                    "tags": ["coverage", "bigwig", "scout", sample_id],
                },
                {
                    "path": Path(
                        mip_dna_analysis_dir, f"{sample_id}_lanes_1234_star_sorted_sj.bed.gz.tbi"
                    ).as_posix(),
                    "archive": False,
                    "tags": ["bed", "scout", "junction", sample_id],
                },
            ]
        )

    return {
        "name": rna_case_id,
        "created": timestamp,
        "expires": timestamp,
        "files": files,
    }


@pytest.fixture(scope="function")
def tomte_analysis_hk_bundle_data(
    rna_case_id: str,
    timestamp: datetime,
    tomte_analysis_dir: Path,
    rna_sample_son_id: str,
    rna_sample_daughter_id: str,
    rna_sample_mother_id: str,
    rna_sample_father_id: str,
) -> dict:
    """Return MIP RNA bundle data for Housekeeper."""

    files: list[dict] = [
        {
            "path": Path(tomte_analysis_dir, f"{rna_case_id}.cram").as_posix(),
            "archive": False,
            "tags": ["cram", rna_case_id],
        },
        {
            "path": Path(tomte_analysis_dir, f"{rna_case_id}_report.selected.pdf").as_posix(),
            "archive": False,
            "tags": ["fusion", "pdf", "clinical", rna_case_id],
        },
        {
            "path": Path(tomte_analysis_dir, f"{rna_case_id}_report.pdf").as_posix(),
            "archive": False,
            "tags": ["fusion", "pdf", "research", rna_case_id],
        },
        {
            "path": Path(tomte_analysis_dir, f"{rna_case_id}_report.pdf").as_posix(),
            "archive": False,
            "tags": ["delivery-report", "rna", rna_case_id],
        },
    ]
    for sample_id in [
        rna_sample_son_id,
        rna_sample_daughter_id,
        rna_sample_mother_id,
        rna_sample_father_id,
    ]:
        files.extend(
            [
                {
                    "path": Path(
                        tomte_analysis_dir, f"{sample_id}_lanes_1_star_sorted_sj.bigWig"
                    ).as_posix(),
                    "archive": False,
                    "tags": ["coverage", "bigwig", "scout", sample_id],
                },
                {
                    "path": Path(
                        tomte_analysis_dir, f"{sample_id}_lanes_1234_star_sorted_sj.bed.gz.tbi"
                    ).as_posix(),
                    "archive": False,
                    "tags": ["bed", "scout", "junction", sample_id],
                },
            ]
        )

    return {
        "name": rna_case_id,
        "created": timestamp,
        "expires": timestamp,
        "files": files,
    }


@pytest.fixture(scope="function")
def balsamic_analysis_hk_bundle_data(
    case_id: str,
    timestamp: datetime,
    balsamic_wgs_analysis_dir: Path,
    sample_id: str,
    snv_vcf_file: str,
    sv_vcf_file: str,
) -> dict:
    """Return Balsamic bundle data for Housekeeper,"""
    return {
        "name": case_id,
        "created": timestamp,
        "expires": timestamp,
        "files": [
            {
                "path": Path(balsamic_wgs_analysis_dir, snv_vcf_file).as_posix(),
                "archive": False,
                "tags": ["vcf-snv-clinical"],
            },
            {
                "path": Path(balsamic_wgs_analysis_dir, sv_vcf_file).as_posix(),
                "archive": False,
                "tags": ["vcf-sv-clinical"],
            },
            {
                "path": Path(balsamic_wgs_analysis_dir, "umi.sv.vcf").as_posix(),
                "archive": False,
                "tags": ["vcf-umi-snv-clinical"],
            },
            {
                "path": Path(balsamic_wgs_analysis_dir, "adm1.cram").as_posix(),
                "archive": False,
                "tags": ["cram", sample_id],
            },
            {
                "path": Path(balsamic_wgs_analysis_dir, "ascat.output.pdf").as_posix(),
                "archive": False,
                "tags": ["ascatngs", "visualization", sample_id],
            },
            {
                "path": Path(balsamic_wgs_analysis_dir, "coverage_qc_report.pdf").as_posix(),
                "archive": False,
                "tags": ["delivery-report"],
            },
        ],
    }


@pytest.fixture(scope="function")
def raredisease_analysis_hk_bundle_data(
    case_id: str,
    sample_id: str,
    timestamp: datetime,
    raredisease_analysis_dir: Path,
    delivery_report_html: Path,
) -> dict:
    """Return RAREDISEASE Housekeeper bundle data."""
    return {
        "name": case_id,
        "created": timestamp,
        "expires": timestamp,
        "files": [
            {
                "path": str(raredisease_analysis_dir / "multiqc.html"),
                "archive": False,
                "tags": [AnalysisTag.MULTIQC_HTML],
            },
            {
                "path": delivery_report_html.as_posix(),
                "archive": False,
                "tags": [HK_DELIVERY_REPORT_TAG],
            },
            {
                "path": str(raredisease_analysis_dir / "snv_vcf_file.vcf"),
                "archive": False,
                "tags": ["vcf-snv-clinical"],
            },
            {
                "path": Path(raredisease_analysis_dir, "sv_vcf_file.vcf").as_posix(),
                "archive": False,
                "tags": ["vcf-sv-clinical"],
            },
            {
                "path": Path(raredisease_analysis_dir, "snv_research_vcf_file.vcf").as_posix(),
                "archive": False,
                "tags": ["vcf-snv-research"],
            },
            {
                "path": Path(raredisease_analysis_dir, "sv_research_vcf_file.vcf").as_posix(),
                "archive": False,
                "tags": ["vcf-sv-research"],
            },
            {
                "path": Path(raredisease_analysis_dir, "str.vcf").as_posix(),
                "archive": False,
                "tags": ["vcf-str"],
            },
            {
                "path": Path(raredisease_analysis_dir, "smn.vcf").as_posix(),
                "archive": False,
                "tags": ["smn-calling"],
            },
            {
                "path": Path(raredisease_analysis_dir, "adm1.cram").as_posix(),
                "archive": False,
                "tags": ["cram", sample_id],
            },
            {
                "path": Path(raredisease_analysis_dir, "report.pdf").as_posix(),
                "archive": False,
                "tags": ["delivery-report"],
            },
            {
                "path": Path(raredisease_analysis_dir, "adm1.mt.bam").as_posix(),
                "archive": False,
                "tags": ["bam-mt", sample_id],
            },
            {
                "path": Path(raredisease_analysis_dir, "vcf2cytosure.txt").as_posix(),
                "archive": False,
                "tags": ["vcf2cytosure", sample_id],
            },
            {
                "path": Path(raredisease_analysis_dir, "multiqc.html").as_posix(),
                "archive": False,
                "tags": ["multiqc-html", sample_id],
            },
            {
                "path": Path(raredisease_analysis_dir, "manifest." + FileFormat.JSON).as_posix(),
                "archive": False,
                "tags": [NFAnalysisTags.MANIFEST],
            },
        ],
    }


@pytest.fixture(scope="function")
def rnafusion_analysis_hk_bundle_data(
    case_id: str, timestamp: datetime, rnafusion_analysis_dir: Path, delivery_report_html: Path
) -> dict:
    """Get some bundle data for housekeeper."""
    return {
        "name": case_id,
        "created": timestamp,
        "expires": timestamp,
        "files": [
            {
                "path": str(rnafusion_analysis_dir / "multiqc.html"),
                "archive": False,
                "tags": ["multiqc-html", "rna"],
            },
            {
                "path": Path(rnafusion_analysis_dir, "rnafusion_report.html").as_posix(),
                "archive": False,
                "tags": ["fusionreport", "research"],
            },
            {
                "path": delivery_report_html.as_posix(),
                "archive": False,
                "tags": [HK_DELIVERY_REPORT_TAG],
            },
        ],
    }


@pytest.fixture
def balsamic_analysis_hk_version(
    housekeeper_api: MockHousekeeperAPI, balsamic_analysis_hk_bundle_data: dict, helpers
) -> MockHousekeeperAPI:
    """Return Housekeeper version for a Balsamic bundle."""
    return helpers.ensure_hk_version(housekeeper_api, balsamic_analysis_hk_bundle_data)


@pytest.fixture
def mip_dna_analysis_hk_version(
    housekeeper_api: MockHousekeeperAPI, mip_dna_analysis_hk_bundle_data: dict, helpers
) -> MockHousekeeperAPI:
    """Return Housekeeper version for a MIP DNA bundle."""
    return helpers.ensure_hk_version(housekeeper_api, mip_dna_analysis_hk_bundle_data)


@pytest.fixture
def raredisease_analysis_hk_version(
    housekeeper_api: MockHousekeeperAPI, raredisease_analysis_hk_bundle_data: dict, helpers
) -> MockHousekeeperAPI:
    """Return Housekeeper version for a MIP DNA bundle."""
    return helpers.ensure_hk_version(housekeeper_api, raredisease_analysis_hk_bundle_data)


@pytest.fixture
def mip_dna_analysis_hk_api(
    housekeeper_api: MockHousekeeperAPI, mip_dna_analysis_hk_bundle_data: dict, helpers
) -> MockHousekeeperAPI:
    """Return a Housekeeper API populated with MIP DNA analysis files."""
    helpers.ensure_hk_version(housekeeper_api, mip_dna_analysis_hk_bundle_data)
    return housekeeper_api


@pytest.fixture
def mip_rna_analysis_hk_api(
    housekeeper_api: MockHousekeeperAPI, mip_rna_analysis_hk_bundle_data: dict, helpers
) -> MockHousekeeperAPI:
    """Return a Housekeeper API populated with MIP RNA analysis files."""
    helpers.ensure_hk_version(housekeeper_api, mip_rna_analysis_hk_bundle_data)
    return housekeeper_api


@pytest.fixture
def tomte_analysis_hk_api(
    housekeeper_api: MockHousekeeperAPI, tomte_analysis_hk_bundle_data: dict, helpers
) -> MockHousekeeperAPI:
    """Return a Housekeeper API populated with MIP RNA analysis files."""
    helpers.ensure_hk_version(housekeeper_api, tomte_analysis_hk_bundle_data)
    return housekeeper_api


@pytest.fixture
def balsamic_analysis_hk_api(
    housekeeper_api: MockHousekeeperAPI, balsamic_analysis_hk_bundle_data: dict, helpers
) -> MockHousekeeperAPI:
    """Return a Housekeeper API populated with Balsamic analysis files."""
    helpers.ensure_hk_version(housekeeper_api, balsamic_analysis_hk_bundle_data)
    return housekeeper_api


@pytest.fixture
def raredisease_analysis_hk_api(
    housekeeper_api: MockHousekeeperAPI, raredisease_analysis_hk_bundle_data: dict, helpers
) -> MockHousekeeperAPI:
    """Return a Housekeeper API populated with some RAREDISEASE analysis files."""
    helpers.ensure_hk_version(housekeeper_api, raredisease_analysis_hk_bundle_data)
    return housekeeper_api


@pytest.fixture
def rnafusion_analysis_hk_api(
    housekeeper_api: MockHousekeeperAPI, rnafusion_analysis_hk_bundle_data: dict, helpers
) -> MockHousekeeperAPI:
    """Return a housekeeper api populated with some rnafusion analysis files"""
    helpers.ensure_hk_version(housekeeper_api, rnafusion_analysis_hk_bundle_data)
    return housekeeper_api


@pytest.fixture
def mip_dna_analysis(
    analysis_store_trio: Store, case_id: str, timestamp: datetime, helpers: StoreHelpers
) -> Analysis:
    """Return a MIP DNA analysis object."""
    helpers.add_synopsis_to_case(store=analysis_store_trio, case_id=case_id)
    case: Case = analysis_store_trio.get_case_by_internal_id(internal_id=case_id)
    analysis: Analysis = helpers.add_analysis(
        store=analysis_store_trio,
        case=case,
        started_at=timestamp,
        completed_at=timestamp,
        workflow=Workflow.MIP_DNA,
    )
    for link in case.links:
        helpers.add_phenotype_groups_to_sample(
            store=analysis_store_trio, sample_id=link.sample.internal_id
        )
        helpers.add_phenotype_terms_to_sample(
            store=analysis_store_trio, sample_id=link.sample.internal_id
        )
        helpers.add_subject_id_to_sample(
            store=analysis_store_trio, sample_id=link.sample.internal_id
        )
    return analysis


@pytest.fixture
def balsamic_analysis_obj(analysis_obj: Analysis) -> Analysis:
    """Return a Balsamic analysis object."""
    analysis_obj.workflow = Workflow.BALSAMIC
    for link_object in analysis_obj.case.links:
        link_object.sample.application_version.application.prep_category = (
            SeqLibraryPrepCategory.WHOLE_EXOME_SEQUENCING
        )
        link_object.case.data_analysis = Workflow.BALSAMIC
    return analysis_obj


@pytest.fixture
def balsamic_umi_analysis_obj(analysis_obj: Analysis) -> Analysis:
    """Return a Balsamic UMI analysis object."""
    analysis_obj.workflow = Workflow.BALSAMIC_UMI
    for link_object in analysis_obj.case.links:
        link_object.sample.application_version.application.prep_category = (
            SeqLibraryPrepCategory.WHOLE_EXOME_SEQUENCING
        )
        link_object.case.data_analysis = Workflow.BALSAMIC_UMI

    return analysis_obj


@pytest.fixture
def raredisease_analysis_obj(analysis_obj: Analysis) -> Analysis:
    """Return a Balsamic analysis object."""
    analysis_obj.workflow = Workflow.RAREDISEASE
    for link_object in analysis_obj.case.links:
        link_object.sample.application_version.application.prep_category = (
            SeqLibraryPrepCategory.WHOLE_GENOME_SEQUENCING
        )
        link_object.sample.reference_genome = GenomeVersion.GRCh37
        link_object.case.data_analysis = Workflow.RAREDISEASE
    return analysis_obj


@pytest.fixture
def rnafusion_analysis_obj(analysis_obj: Analysis) -> Analysis:
    """Return a RNAfusion analysis object."""
    analysis_obj.workflow = Workflow.RNAFUSION
    for link_object in analysis_obj.case.links:
        link_object.sample.application_version.application.prep_category = (
            SeqLibraryPrepCategory.WHOLE_TRANSCRIPTOME_SEQUENCING
        )
        link_object.case.data_analysis = Workflow.RNAFUSION
        link_object.case.panels = None
    return analysis_obj


@pytest.fixture
def mip_config_builder(
    mip_dna_analysis_hk_version: Version,
    mip_dna_analysis: Analysis,
    lims_api: MockLimsAPI,
    mip_analysis_api: MockMipAnalysis,
    madeline_api: MockMadelineAPI,
) -> MipConfigBuilder:
    """Return a MIP config builder."""
    return MipConfigBuilder(
        hk_version_obj=mip_dna_analysis_hk_version,
        analysis_obj=mip_dna_analysis,
        lims_api=lims_api,
        mip_analysis_api=mip_analysis_api,
        madeline_api=madeline_api,
    )


@pytest.fixture
def raredisease_config_builder(
    raredisease_analysis_hk_version: Version,
    raredisease_analysis_obj: Analysis,
    lims_api: MockLimsAPI,
    raredisease_analysis_api: RarediseaseAnalysisAPI,
    madeline_api: MockMadelineAPI,
) -> MipConfigBuilder:
    """Return a MIP config builder."""
    return RarediseaseConfigBuilder(
        hk_version_obj=raredisease_analysis_hk_version,
        analysis_obj=raredisease_analysis_obj,
        lims_api=lims_api,
        madeline_api=madeline_api,
        raredisease_analysis_api=raredisease_analysis_api,
    )


@pytest.fixture
def balsamic_config_builder(
    balsamic_analysis_hk_version: Version,
    balsamic_analysis_obj: Analysis,
    lims_api: MockLimsAPI,
) -> BalsamicConfigBuilder:
    """Return a Balsamic config builder."""
    return BalsamicConfigBuilder(
        hk_version_obj=balsamic_analysis_hk_version,
        analysis_obj=balsamic_analysis_obj,
        lims_api=lims_api,
    )


@pytest.fixture
def mip_load_config(
    mip_dna_analysis_dir: Path,
    case_id: str,
    customer_id: str,
    snv_vcf_file: str,
    delivery_report_html: Path,
) -> MipLoadConfig:
    """Return a valid MIP load_config."""
    return MipLoadConfig(
        owner=customer_id,
        family=case_id,
        vcf_snv=Path(mip_dna_analysis_dir, snv_vcf_file).as_posix(),
        track=UploadTrack.RARE_DISEASE.value,
        delivery_report=delivery_report_html.as_posix(),
    )


@pytest.fixture
def lims_api(lims_samples: list[dict]) -> MockLimsAPI:
    """Return a LIMS API."""
    return MockLimsAPI(samples=lims_samples)


@pytest.fixture
def mip_analysis_api(cg_context: CGConfig) -> MockMipAnalysis:
    """Return a MIP analysis API."""
    return MockMipAnalysis(config=cg_context, workflow=Workflow.MIP_DNA)


@pytest.fixture
def upload_scout_api(
    cg_context: CGConfig,
    scout_api: MockScoutAPI,
    madeline_api: MockMadelineAPI,
    lims_samples: list[dict],
    housekeeper_api: MockHousekeeperAPI,
    store: Store,
) -> UploadScoutAPI:
    """Return upload Scout API."""
    analysis_mock = MockMipAnalysis(config=cg_context, workflow=Workflow.MIP_DNA)
    lims_api = MockLimsAPI(samples=lims_samples)

    return UploadScoutAPI(
        hk_api=housekeeper_api,
        scout_api=scout_api,
        madeline_api=madeline_api,
        analysis_api=analysis_mock,
        lims_api=lims_api,
        status_db=store,
    )


@pytest.fixture
def upload_mip_analysis_scout_api(
    cg_context: CGConfig,
    scout_api: MockScoutAPI,
    madeline_api: MockMadelineAPI,
    lims_samples: list[dict],
    mip_dna_analysis_hk_api: MockHousekeeperAPI,
    store: Store,
) -> Generator[UploadScoutAPI, None, None]:
    """Return MIP upload Scout API."""
    analysis_mock = MockMipAnalysis(config=cg_context, workflow=Workflow.MIP_DNA)
    lims_api = MockLimsAPI(samples=lims_samples)

    yield UploadScoutAPI(
        hk_api=mip_dna_analysis_hk_api,
        scout_api=scout_api,
        madeline_api=madeline_api,
        analysis_api=analysis_mock,
        lims_api=lims_api,
        status_db=store,
    )


@pytest.fixture
def upload_tomte_analysis_scout_api(
    cg_context: CGConfig,
    scout_api: MockScoutAPI,
    madeline_api: MockMadelineAPI,
    lims_samples: list[dict],
    tomte_analysis_hk_api: MockHousekeeperAPI,
    store: Store,
) -> Generator[UploadScoutAPI, None, None]:
    """Return MIP upload Scout API."""
    analysis_mock = MockMipAnalysis(config=cg_context, workflow=Workflow.TOMTE)
    lims_api = MockLimsAPI(samples=lims_samples)

    yield UploadScoutAPI(
        hk_api=tomte_analysis_hk_api,
        scout_api=scout_api,
        madeline_api=madeline_api,
        analysis_api=analysis_mock,
        lims_api=lims_api,
        status_db=store,
    )


@pytest.fixture
def upload_balsamic_analysis_scout_api(
    cg_context: CGConfig,
    scout_api: MockScoutAPI,
    madeline_api: MockMadelineAPI,
    lims_samples: list[dict],
    balsamic_analysis_hk_api: MockHousekeeperAPI,
    analysis_store: Store,
) -> Generator[UploadScoutAPI, None, None]:
    """Return Balsamic upload Scout API."""
    analysis_mock = MockBalsamicAnalysis(config=cg_context, workflow=Workflow.BALSAMIC)
    lims_api = MockLimsAPI(samples=lims_samples)

    yield UploadScoutAPI(
        hk_api=balsamic_analysis_hk_api,
        scout_api=scout_api,
        madeline_api=madeline_api,
        analysis_api=analysis_mock,
        lims_api=lims_api,
        status_db=analysis_store,
    )


@pytest.fixture
def upload_raredisease_analysis_scout_api(
    raredisease_context: CGConfig,
    scout_api: MockScoutAPI,
    madeline_api: MockMadelineAPI,
    lims_samples: list[dict],
    raredisease_analysis_hk_api: MockHousekeeperAPI,
    analysis_store: Store,
) -> Generator[UploadScoutAPI, None, None]:
    """Return RAREDISEASE upload Scout API."""
    lims_api = MockLimsAPI(samples=lims_samples)

    yield UploadScoutAPI(
        hk_api=raredisease_analysis_hk_api,
        scout_api=scout_api,
        madeline_api=madeline_api,
        analysis_api=raredisease_context.meta_apis["analysis_api"],
        lims_api=lims_api,
        status_db=analysis_store,
    )


@pytest.fixture
def raredisease_analysis_api(cg_context: CGConfig) -> NextflowAnalysis:
    """Return a RAREDISEASE analysis API."""
    return NextflowAnalysis(config=cg_context, workflow=Workflow.RAREDISEASE, sample_metrics=dict())


@pytest.fixture
def rna_dna_sample_case_map(
    rna_sample_son_id: str,
    rna_store: Store,
    upload_scout_api: UploadScoutAPI,
) -> dict[str, list[str]]:
    """Return a valid RNA-DNA case map."""
    rna_sample: Sample = rna_store.get_sample_by_internal_id(internal_id=rna_sample_son_id)

    # WHEN adding the RNA sample rna_dna_case_map
    rna_dna_sample_case_map: dict[str, dict[str, list[str]]] = {}
    upload_scout_api._map_dna_samples_related_to_rna_sample(
        rna_sample=rna_sample, rna_dna_sample_case_map=rna_dna_sample_case_map
    )

    return rna_dna_sample_case_map


@pytest.fixture
def upload_rnafusion_analysis_scout_api(
    cg_context: CGConfig,
    scout_api: MockScoutAPI,
    madeline_api: MockMadelineAPI,
    lims_samples: list[dict],
    rnafusion_analysis_hk_api: MockHousekeeperAPI,
    analysis_store: Store,
) -> UploadScoutAPI:
    """Fixture for upload_scout_api."""
    analysis_mock = MockMipAnalysis(config=cg_context, workflow=Workflow.MIP_DNA)
    lims_api = MockLimsAPI(samples=lims_samples)

    _api = UploadScoutAPI(
        hk_api=rnafusion_analysis_hk_api,
        scout_api=scout_api,
        madeline_api=madeline_api,
        analysis_api=analysis_mock,
        lims_api=lims_api,
        status_db=analysis_store,
    )

    return _api



from enum import StrEnum
from pathlib import Path

from pydantic import field_validator

from cg.constants.constants import GenomeVersion, Strandedness
from cg.constants.sample_sources import SourceType
from cg.models.nf_analysis import NextflowSampleSheetEntry, WorkflowParameters
from cg.models.qc_metrics import QCMetrics
from cg.utils.utils import replace_non_alphanumeric


class TomteSampleSheetEntry(NextflowSampleSheetEntry):
    """Tomte sample model is used when building the sample sheet."""

    case_id: str
    strandedness: Strandedness

    @property
    def reformat_sample_content(self) -> list[list[str]]:
        """Reformat sample sheet content as a list of lists, where
        each list represents a line in the final file."""
        return [
            [
                self.case_id,
                self.name,
                fastq_forward_read_path,
                fastq_reverse_read_path,
                str(self.strandedness),
            ]
            for fastq_forward_read_path, fastq_reverse_read_path in zip(
                self.fastq_forward_read_paths, self.fastq_reverse_read_paths
            )
        ]


class TomteSampleSheetHeaders(StrEnum):
    case_id: str = "case"
    name: str = "sample"
    fastq_1: str = "fastq_1"
    fastq_2: str = "fastq_2"
    strandedness: str = "strandedness"

    @classmethod
    def list(cls) -> list[str]:
        return list(map(lambda header: header.value, cls))


class TomteParameters(WorkflowParameters):
    """Model for Tomte parameters."""

    gene_panel_clinical_filter: Path
    tissue: str
    genome: str = GenomeVersion.HG38

    @field_validator("tissue", mode="before")
    @classmethod
    def restrict_tissue_values(cls, tissue: str | None) -> str:
        if tissue:
            return replace_non_alphanumeric(string=tissue)
        else:
            return SourceType.UNKNOWN

    @field_validator("genome", mode="before")
    @classmethod
    def restrict_genome_values(cls, genome: str) -> str:
        if genome == GenomeVersion.HG38:
            return GenomeVersion.GRCh38.value
        elif genome == GenomeVersion.HG19:
            return GenomeVersion.GRCh37.value


class TomteQCMetrics(QCMetrics):
    """Tomte QC metrics."""

    after_filtering_gc_content: float
    after_filtering_q20_rate: float
    after_filtering_q30_rate: float
    after_filtering_read1_mean_length: float
    before_filtering_total_reads: float
    median_5prime_to_3prime_bias: float
    pct_adapter: float
    pct_duplication: float
    pct_intergenic_bases: float
    pct_intronic_bases: float
    pct_mrna_bases: float
    pct_ribosomal_bases: float
    pct_surviving: float
    uniquely_mapped_percent: float



"""Tests the find business data part of the Cg store API related to sample model."""

from typing import Any

import pytest
from _pytest.fixtures import FixtureRequest
from sqlalchemy.orm import Query

from cg.constants.sequencing import SeqLibraryPrepCategory
from cg.store.models import Customer, Invoice, Sample
from cg.store.store import Store
from tests.store_helpers import StoreHelpers


def test_get_all_pools_and_samples_for_invoice_by_invoice_id(store: Store, helpers: StoreHelpers):
    """Test that all pools and samples for an invoice can be fetched."""

    # GIVEN a database with a pool and a sample
    pool = helpers.ensure_pool(store=store, name="pool_1")
    sample = helpers.add_sample(store=store, name="sample_1")

    # AND an invoice with the pool and sample
    invoice: Invoice = helpers.ensure_invoice(store=store, pools=[pool], samples=[sample])

    # ASSERT that there is an invoice with a pool and a sample
    assert len(invoice.pools) == 1
    assert len(invoice.samples) == 1

    # WHEN fetching all pools and samples for the invoice
    records = store.get_pools_and_samples_for_invoice_by_invoice_id(invoice_id=invoice.id)
    # THEN the pool and sample should be returned
    assert pool in records
    assert sample in records


def test_get_samples_by_customer_and_subject_id_query(
    store_with_samples_subject_id_and_tumour_status: Store,
    cust123: str,
    test_subject: str,
):
    """Test that samples can be fetched by subject id."""
    # GIVEN a database with two samples that have a subject ID but only one is tumour

    # GIVEN that there are two samples in the store
    assert len(store_with_samples_subject_id_and_tumour_status._get_query(table=Sample).all()) == 2

    # GIVEN that there is a customer with the given customer id
    assert store_with_samples_subject_id_and_tumour_status.get_customer_by_internal_id(
        customer_internal_id=cust123
    )

    # WHEN fetching the sample by subject id and customer_id
    samples = store_with_samples_subject_id_and_tumour_status._get_samples_by_customer_and_subject_id_query(
        subject_id=test_subject, customer_internal_id=cust123
    )

    # THEN two samples should be returned
    assert samples.count() == 2
    # THEN the fetched samples have the subject id used for filtering
    assert all(fetched_sample.subject_id == test_subject for fetched_sample in samples.all())


def test_get_samples_by_customer_and_subject_id_query_missing_subject_id(
    store_with_samples_and_tumour_status_missing_subject_id: Store,
    cust123: str,
    test_subject: str,
):
    """Test that samples can be fetched by subject id."""
    # GIVEN a database with two samples that have a subject ID but only one is tumour

    # GIVEN that there are two samples in the store
    assert (
        len(store_with_samples_and_tumour_status_missing_subject_id._get_query(table=Sample).all())
        == 2
    )

    # GIVEN that there is a customer with the given customer id
    assert store_with_samples_and_tumour_status_missing_subject_id.get_customer_by_internal_id(
        customer_internal_id=cust123
    )

    # WHEN fetching the sample by subject id and customer_id
    samples = store_with_samples_and_tumour_status_missing_subject_id._get_samples_by_customer_and_subject_id_query(
        subject_id=test_subject, customer_internal_id=cust123
    )

    # THEN two samples should be returned
    assert samples.count() == 0


def test_get_samples_by_subject_id(
    store_with_samples_subject_id_and_tumour_status: Store,
    cust123: str,
    test_subject: str,
):
    """Test that samples can be fetched by subject id."""
    # GIVEN a database with two samples that have a subject ID but only one is tumour

    # ASSERT that there are two samples in the store
    assert len(store_with_samples_subject_id_and_tumour_status._get_query(table=Sample).all()) == 2

    # ASSERT that there is a customer with the given customer id
    assert store_with_samples_subject_id_and_tumour_status.get_customer_by_internal_id(
        customer_internal_id=cust123
    )

    # WHEN fetching the sample by subject id and customer_id
    samples = (
        store_with_samples_subject_id_and_tumour_status.get_samples_by_customer_and_subject_id(
            subject_id=test_subject, customer_internal_id=cust123
        )
    )

    # THEN two samples should be returned
    assert samples and len(samples) == 2


def test_get_samples_by_customer_id_list_and_subject_id_and_is_tumour(
    store_with_samples_customer_id_and_subject_id_and_tumour_status: Store,
    customer_ids: list[int] = [1, 2],
    subject_id: str = "test_subject",
    is_tumour: bool = True,
):
    """Test that samples can be fetched by customer ID, subject ID, and tumour status."""
    # GIVEN a database with four samples, two with customer ID 1 and two with customer ID 2

    # ASSERT that there are customers with the given customer IDs
    for customer_id in customer_ids:
        assert store_with_samples_customer_id_and_subject_id_and_tumour_status.get_customer_by_internal_id(
            customer_internal_id=str(customer_id)
        )

    # WHEN fetching the samples by customer ID list, subject ID, and tumour status
    samples = store_with_samples_customer_id_and_subject_id_and_tumour_status.get_samples_by_customer_ids_and_subject_id_and_is_tumour(
        customer_ids=customer_ids, subject_id=subject_id, is_tumour=is_tumour
    )

    # THEN two samples should be returned, one for each customer ID, with the specified subject ID and tumour status
    assert isinstance(samples, list)
    assert len(samples) == 2

    for sample in samples:
        assert isinstance(sample, Sample)

    for customer_id, sample in zip(customer_ids, samples):
        assert sample.customer_id == customer_id
        assert sample.subject_id == subject_id
        assert sample.is_tumour == is_tumour


def test_get_samples_by_customer_id_list_and_subject_id_and_is_tumour_with_non_existing_customer_id(
    store_with_samples_customer_id_and_subject_id_and_tumour_status: Store,
):
    """Test that no samples are returned when filtering on non-existing customer ID."""
    # GIVEN a database with four samples, two with customer ID 1 and two with customer ID 2

    # ASSERT that there are no customers with the given customer IDs
    customer_ids = [1, 2, 3]
    for customer_id in customer_ids:
        if customer_id == 3:
            assert (
                store_with_samples_customer_id_and_subject_id_and_tumour_status.get_customer_by_internal_id(
                    customer_internal_id=str(customer_id)
                )
                is None
            )
        else:
            assert store_with_samples_customer_id_and_subject_id_and_tumour_status.get_customer_by_internal_id(
                customer_internal_id=str(customer_id)
            )

    # WHEN fetching the samples by customer ID list, subject ID, and tumour status
    non_existing_customer_id = [3]
    samples = store_with_samples_customer_id_and_subject_id_and_tumour_status.get_samples_by_customer_ids_and_subject_id_and_is_tumour(
        customer_ids=non_existing_customer_id, subject_id="test_subject", is_tumour=True
    )

    # THEN no samples should be returned
    assert isinstance(samples, list)
    assert len(samples) == 0


def test_get_sample_by_name(store_with_samples_that_have_names: Store, name="test_sample_1"):
    """Test that samples can be fetched by name."""
    # GIVEN a database with two samples of which one has a name

    # ASSERT that there are two samples in the store
    assert len(store_with_samples_that_have_names._get_query(table=Sample).all()) == 4

    # WHEN fetching the sample by name
    samples: Sample = store_with_samples_that_have_names.get_sample_by_name(name=name)

    # THEN one sample should be returned
    assert samples and samples.name == name


def test_has_active_cases_for_sample_analyze(
    store_with_active_sample_analyze: Store,
    sample_internal_id: str = "test_sample_internal_id",
):
    """Test that a sample is active."""
    # GIVEN a store with an active case

    # THEN the sample should be active
    assert (
        store_with_active_sample_analyze.has_active_cases_for_sample(internal_id=sample_internal_id)
        is True
    )


def test_has_active_cases_for_sample_running(
    store_with_active_sample_running: Store,
    sample_internal_id: str = "test_sample_internal_id",
):
    """Test that a sample is active."""
    # GIVEN a store with an active case

    # THEN the sample should be active
    assert (
        store_with_active_sample_running.has_active_cases_for_sample(internal_id=sample_internal_id)
        is True
    )


def test_get_samples_by_customer_and_name(
    store_with_samples_that_have_names: Store,
    name: str = "test_sample_1",
    customer_id="cust000",
):
    """Test that samples can be fetched by name."""
    # GIVEN a database with samples
    # WHEN getting a customer from the store
    customer: Customer = store_with_samples_that_have_names.get_customer_by_internal_id(
        customer_internal_id=customer_id
    )
    assert customer
    # WHEN fetching the sample by name
    sample: Sample = store_with_samples_that_have_names.get_sample_by_customer_and_name(
        customer_entry_id=[customer.id], sample_name=name
    )

    # THEN one sample should be returned
    assert sample
    assert sample.name == name
    assert sample.customer == customer


def test_get_samples_by_customer_and_name_invalid_customer(
    store_with_samples_that_have_names: Store,
    name: str = "test_sample_1",
    customer_id="unrelated_customer",
):
    """Test that samples can be fetched by name does not return a sample when invalid customer is supplied."""
    # GIVEN a database with two samples

    # WHEN getting a customer from the store
    customer: Customer = store_with_samples_that_have_names.get_customer_by_internal_id(
        customer_internal_id=customer_id
    )
    assert customer
    # WHEN fetching the sample by name
    sample: Sample = store_with_samples_that_have_names.get_sample_by_customer_and_name(
        customer_entry_id=[customer.id], sample_name=name
    )

    # THEN one sample should be returned
    assert not sample


def test_get_samples_by_any_id_not_an_attribute_fails(
    store_with_a_sample_that_has_many_attributes_and_one_without: Store,
    identifiers: dict[str, Any] = {
        "non-existent-attribute": "not-an-attribute",
        "non-existent-value": "not-a-value",
    },
):
    """Test that using an attribute not from Sample raises an error."""
    # GIVEN a store with samples
    sample_query: Query = store_with_a_sample_that_has_many_attributes_and_one_without._get_query(
        table=Sample
    )
    assert sample_query.count() > 0

    # WHEN trying to filter using an attribute not of Sample
    with pytest.raises(AttributeError) as exc_info:
        store_with_a_sample_that_has_many_attributes_and_one_without.get_samples_by_any_id(
            **identifiers
        )

    # THEN the error message should contain the non-existent-attribute
    assert str(exc_info.value) == "type object 'Sample' has no attribute 'non-existent-attribute'"


def test_get_samples_by_any_id_exclusive_filtering_gives_empty_query(
    store_with_a_sample_that_has_many_attributes_and_one_without: Store,
):
    """Test that using mutually exclusive filtering conditions give an empty query."""
    # GIVEN a store with two samples
    sample_query: Query = store_with_a_sample_that_has_many_attributes_and_one_without._get_query(
        table=Sample
    )
    assert sample_query.count() == 2

    # GIVEN that the samples in the query have different values of two attributes
    sample_1: Sample = sample_query[0]
    sample_2: Sample = sample_query[1]
    assert sample_1.name != sample_2.name
    assert sample_1.is_tumour != sample_2.is_tumour

    # WHEN filtering twice mutually exclusive conditions
    identifiers: dict[str, str] = {
        "name": sample_1.name,
        "is_tumour": sample_2.is_tumour,
    }
    filtered_query: Query = (
        store_with_a_sample_that_has_many_attributes_and_one_without.get_samples_by_any_id(
            **identifiers
        )
    )

    # THEN the filtered query is empty
    assert filtered_query.count() == 0


def test_samples_to_receive_external(sample_store: Store, helpers: StoreHelpers):
    """Test fetching external sample."""
    # GIVEN a store with a mixture of samples
    assert len(sample_store._get_query(table=Sample).all()) > 1

    # WHEN finding external samples to receive
    samples: list[Sample] = sample_store.get_samples_to_receive(external=True)

    # THEN samples should be a list of samples
    assert isinstance(samples, list)
    assert isinstance(samples[0], Sample)

    # THEN assert that only the external sample is returned
    assert len(samples) == 1

    first_sample = samples[0]
    # THEN assert that the sample is external in database
    assert first_sample.application_version.application.is_external is True
    # THEN assert that the sample is does not have a received at stamp
    assert first_sample.received_at is None


def test_get_samples_to_receive_internal(sample_store):
    """Test fetching internal samples."""
    # GIVEN a store with samples in a mix of states
    assert len(sample_store._get_query(table=Sample).all()) > 1
    assert (
        len(
            [sample for sample in sample_store._get_query(table=Sample).all() if sample.received_at]
        )
        > 1
    )

    # WHEN finding which samples are in queue to receive
    assert len(sample_store.get_samples_to_receive()) == 3
    first_sample = sample_store.get_samples_to_receive()[0]

    # THEN samples should be a sample
    assert isinstance(first_sample, Sample)

    assert first_sample.application_version.application.is_external is False
    assert first_sample.received_at is None


def test_samples_to_sequence(sample_store):
    """Test fetching samples to sequence."""
    # GIVEN a store with sample in a mix of states
    assert len(sample_store._get_query(table=Sample).all()) > 1
    assert (
        len(
            [
                sample
                for sample in sample_store._get_query(table=Sample).all()
                if sample.last_sequenced_at
            ]
        )
        >= 1
    )

    # WHEN finding which samples are in queue to be sequenced
    sequence_samples: list[Sample] = sample_store.get_samples_to_sequence()

    # THEN samples should be a list of samples
    assert isinstance(sequence_samples, list)
    assert isinstance(sequence_samples[0], Sample)

    # THEN it should list the received and partly sequenced samples
    assert len(sequence_samples) == 2
    assert {sample.name for sample in sequence_samples} == set(
        ["sequenced-partly", "received-prepared"]
    )
    for sample in sequence_samples:
        assert sample.last_sequenced_at is None
        if sample.name == "sequenced-partly":
            assert sample.reads > 0


def test_samples_to_prepare(sample_store):
    """Test fetching samples to prepare."""
    # GIVEN a store with sample in a mix of states
    assert len(sample_store._get_query(table=Sample).all()) > 1
    assert (
        len(
            [sample for sample in sample_store._get_query(table=Sample).all() if sample.prepared_at]
        )
        >= 1
    )

    # WHEN finding which samples are in queue to be prepared
    prepare_samples: list[Sample] = sample_store.get_samples_to_prepare()

    # THEN samples should be a list of samples
    assert isinstance(prepare_samples, list)
    assert isinstance(prepare_samples[0], Sample)

    # THEN it should list the received sample
    assert len(prepare_samples) == 1
    assert prepare_samples[0].name == "received"


def test_get_sample_by_entry_id(sample_store, entry_id=1):
    """Test fetching a sample by entry id."""
    # GIVEN a store with a sample
    assert len(sample_store._get_query(table=Sample).all()) > 1

    # WHEN finding a sample by entry id
    sample: Sample = sample_store.get_sample_by_entry_id(entry_id=entry_id)

    # THEN samples should be a list of samples
    assert isinstance(sample, Sample)

    # THEN it should return the sample
    assert sample.id == entry_id


def test_get_sample_by_internal_id(sample_store, internal_id="test_internal_id"):
    """Test fetching a sample by internal id."""
    # GIVEN a store with a sample
    assert len(sample_store._get_query(table=Sample).all()) > 1

    # WHEN finding a sample by internal id
    sample: Sample = sample_store.get_sample_by_internal_id(internal_id=internal_id)

    # THEN samples should be a list of samples
    assert isinstance(sample, Sample)

    # THEN it should return the sample
    assert sample.internal_id == internal_id


@pytest.mark.parametrize(
    "object_type, identifier_fixture",
    [
        ("sample", "sample_id_sequenced_on_multiple_flow_cells"),
        ("flow_cell", "novaseq_x_flow_cell_id"),
        ("case", "case_id_for_sample_on_multiple_flow_cells"),
    ],
    ids=["sample", "flow_cell", "case"],
)
def test_get_samples_by_identifier(
    re_sequenced_sample_illumina_data_store: Store,
    object_type: str,
    identifier_fixture: str,
    request: FixtureRequest,
):
    """Test that samples are returned for any instance of an identifier."""
    # GIVEN a store with samples, an identifier and an object type
    store: Store = re_sequenced_sample_illumina_data_store
    identifier: str = request.getfixturevalue(identifier_fixture)

    # WHEN fetching the samples by identifier
    samples: list[Sample] = store.get_samples_by_identifier(
        object_type=object_type, identifier=identifier
    )

    # THEN a list of samples should be returned
    assert isinstance(samples, list)
    assert isinstance(samples[0], Sample)


def test_get_samples_to_deliver(sample_store):
    """Test fetching samples to deliver."""
    # GIVEN a store with a sample
    assert len(sample_store._get_query(table=Sample).all()) > 1

    # WHEN finding samples to deliver
    samples = sample_store.get_samples_to_deliver()

    # THEN samples should be a list of samples
    assert isinstance(samples, list)
    assert isinstance(samples[0], Sample)

    # THEN it should return the samples that are sequenced but not delivered
    assert len(samples) == 2
    assert {sample.name for sample in samples} == set(["to-deliver", "sequenced"])


def test_get_samples_to_invoice_query(sample_store):
    """Test fetching samples to invoice."""
    # GIVEN a store with a sample
    assert len(sample_store._get_query(table=Sample).all()) > 1

    # WHEN finding samples to invoice
    sample = sample_store.get_samples_to_invoice_query().first()

    # THEN samples should be a list of samples
    assert isinstance(sample, Sample)

    # THEN it should return all samples that are not invoiced
    assert sample
    assert sample.name == "delivered"


def test_get_samples_not_invoiced(sample_store):
    """Test getting samples not invoiced."""
    # GIVEN a store with a sample
    assert len(sample_store._get_query(table=Sample).all()) > 1

    # WHEN finding samples to invoice
    samples = sample_store.get_samples_not_invoiced()

    # THEN samples should be a list of samples
    assert isinstance(samples, list)
    assert isinstance(samples[0], Sample)

    # THEN it should return all samples that are not invoiced
    assert len(samples) == len(sample_store._get_query(table=Sample).all())


def test_get_samples_not_down_sampled(sample_store: Store, helpers: StoreHelpers, sample_id: int):
    """Test getting samples not down sampled."""
    # GIVEN a store with a sample
    assert len(sample_store._get_query(table=Sample).all()) > 1

    # WHEN finding samples to invoice
    samples = sample_store.get_samples_not_down_sampled()

    # THEN samples should be a list of samples
    assert isinstance(samples, list)
    assert isinstance(samples[0], Sample)

    # THEN it should return all samples in the store
    assert len(samples) == len(sample_store._get_query(table=Sample).all())


def test_get_samples_to_invoice_for_customer(
    store_with_samples_for_multiple_customers: Store,
    helpers: StoreHelpers,
    three_customer_ids: list[str],
):
    """Test that samples to invoice can be returned for a customer."""
    # GIVEN a database with samples for a customer

    # THEN the one customer can be retrieved
    customer: Customer = store_with_samples_for_multiple_customers.get_customer_by_internal_id(
        customer_internal_id=three_customer_ids[1]
    )
    assert customer

    # WHEN getting the samples to invoice for a customer
    samples: list[Sample] = (
        store_with_samples_for_multiple_customers.get_samples_to_invoice_for_customer(
            customer=customer,
        )
    )

    # THEN the samples should be returned
    assert samples
    assert len(samples) == 1

    assert samples[0].customer.internal_id == three_customer_ids[1]


def test_get_samples_by_customer_id_and_pattern_with_collaboration(
    store_with_samples_for_multiple_customers: Store,
    helpers: StoreHelpers,
    three_customer_ids: list[str],
):
    """Test that samples can be returned for a customer."""
    # GIVEN a database with samples for a customer

    # THEN the one customer can be retrieved
    customer: set[Customer] = store_with_samples_for_multiple_customers.get_customer_by_internal_id(
        customer_internal_id=three_customer_ids[1]
    ).collaborators
    assert customer

    # WHEN getting the samples for a customer
    samples: list[Sample] = (
        store_with_samples_for_multiple_customers.get_samples_by_customers_and_pattern(
            customers=customer,
            pattern="sample",
        )
    )

    # THEN the samples should be returned
    assert samples
    assert len(samples) == 3

    for sample in samples:
        assert "sample" in sample.name


def test_get_related_samples(
    store_with_rna_and_dna_samples_and_cases: Store,
    rna_sample: Sample,
    related_dna_samples: list[Sample],
    rna_sample_collaborators: set[Customer],
):
    # GIVEN a database with an RNA sample and several DNA samples with the same subject_id and tumour status as the given sample
    # GIVEN that all customers are in a collaboration
    # GIVEN a list of dna prep categories
    dna_prep_categories: list[SeqLibraryPrepCategory] = [
        SeqLibraryPrepCategory.WHOLE_GENOME_SEQUENCING,
        SeqLibraryPrepCategory.TARGETED_GENOME_SEQUENCING,
        SeqLibraryPrepCategory.WHOLE_EXOME_SEQUENCING,
    ]

    # WHEN getting the related DNA samples to the given sample
    fetched_related_dna_samples = store_with_rna_and_dna_samples_and_cases.get_related_samples(
        sample_internal_id=rna_sample.internal_id,
        prep_categories=dna_prep_categories,
        collaborators=rna_sample_collaborators,
    )

    # THEN the correct set of samples is returned
    assert set(related_dna_samples) == set(fetched_related_dna_samples)



"""Delivery report CG context fixtures."""

from datetime import datetime
from pathlib import Path

import click
import pytest
from pytest_mock import MockFixture

from cg.apps.lims import LimsAPI
from cg.cli.generate.delivery_report.base import generate_delivery_report
from cg.clients.chanjo2.models import CoverageMetrics
from cg.constants import DataDelivery, Workflow
from cg.meta.delivery_report.delivery_report_api import DeliveryReportAPI
from cg.meta.delivery_report.raredisease import RarediseaseDeliveryReportAPI
from cg.meta.delivery_report.rnafusion import RnafusionDeliveryReportAPI
from cg.meta.workflow.analysis import AnalysisAPI
from cg.meta.workflow.nf_analysis import NfAnalysisAPI
from cg.meta.workflow.raredisease import RarediseaseAnalysisAPI
from cg.meta.workflow.rnafusion import RnafusionAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.models.delivery_report.report import ScoutVariantsFiles
from cg.store.models import Case, Sample
from cg.store.store import Store
from tests.conftest import raredisease_case_id
from tests.store_helpers import StoreHelpers


@pytest.fixture
def delivery_report_context(
    capture_kit: str,
    cg_context: CGConfig,
    libary_sequencing_method: str,
    library_prep_method: str,
    lims_samples: list[dict],
    mocker: MockFixture,
) -> CGConfig:
    """Delivery report generation context."""
    mocker.patch.object(AnalysisAPI, "get_gene_ids_from_scout", return_value=[])
    mocker.patch.object(DeliveryReportAPI, "get_delivery_report_from_hk", return_value=None)
    mocker.patch.object(LimsAPI, "sample", return_value=lims_samples[0])
    mocker.patch.object(LimsAPI, "get_prep_method", return_value=library_prep_method)
    mocker.patch.object(LimsAPI, "get_sequencing_method", return_value=libary_sequencing_method)
    mocker.patch.object(LimsAPI, "capture_kit", return_value=capture_kit)
    mocker.patch.object(LimsAPI, "get_sample_dv200", return_value=75.0)
    mocker.patch.object(LimsAPI, "get_latest_rna_input_amount", return_value=300.0)
    mocker.patch.object(LimsAPI, "get_sample_rin", return_value=10)
    return cg_context


@pytest.fixture
def raredisease_delivery_report_store_context(
    base_store: Store,
    helpers: StoreHelpers,
    raredisease_case_id: str,
    sample_id: str,
    ticket_id: str,
    total_sequenced_reads_pass: int,
    wgs_application_tag: str,
) -> Store:
    """Raredisease delivery report Statusdb fixture."""
    case: Case = helpers.add_case(
        store=base_store,
        internal_id=raredisease_case_id,
        name=raredisease_case_id,
        data_analysis=Workflow.RAREDISEASE,
    )
    sample: Sample = helpers.add_sample(
        store=base_store,
        internal_id=sample_id,
        last_sequenced_at=datetime.now(),
        received_at=datetime.now(),
        reads=total_sequenced_reads_pass,
        application_tag=wgs_application_tag,
        original_ticket=ticket_id,
        reference_genome="hg19",
    )
    helpers.add_relationship(store=base_store, case=case, sample=sample)
    helpers.add_analysis(
        store=base_store,
        case=case,
        started_at=datetime.now(),
        completed_at=datetime.now(),
        workflow=Workflow.RAREDISEASE,
        data_delivery=DataDelivery.ANALYSIS_SCOUT,
    )
    return base_store


@pytest.fixture
def rnafusion_delivery_report_store_context(
    base_store: Store,
    helpers: StoreHelpers,
    rnafusion_case_id: str,
    sample_id: str,
    ticket_id: str,
    total_sequenced_reads_pass: int,
    wts_application_tag: str,
) -> Store:
    """Rnafusion delivery report Statusdb fixture."""
    case: Case = helpers.add_case(
        store=base_store,
        internal_id=rnafusion_case_id,
        name=rnafusion_case_id,
        data_analysis=Workflow.RNAFUSION,
    )
    sample: Sample = helpers.add_sample(
        store=base_store,
        internal_id=sample_id,
        last_sequenced_at=datetime.now(),
        received_at=datetime.now(),
        reads=total_sequenced_reads_pass,
        application_tag=wts_application_tag,
        original_ticket=ticket_id,
    )
    helpers.add_relationship(store=base_store, case=case, sample=sample)
    helpers.add_analysis(
        store=base_store,
        case=case,
        started_at=datetime.now(),
        completed_at=datetime.now(),
        workflow=Workflow.RNAFUSION,
        data_delivery=DataDelivery.ANALYSIS_SCOUT,
    )
    return base_store


@pytest.fixture
def raredisease_delivery_report_context(
    coverage_metrics: CoverageMetrics,
    delivery_report_context: CGConfig,
    mocker: MockFixture,
    raredisease_delivery_report_store_context: Store,
    raredisease_multiqc_json_metrics_path: Path,
    scout_variants_files: ScoutVariantsFiles,
) -> CGConfig:
    """Raredisease context for the delivery report generation."""

    # Raredisease delivery report context
    delivery_report_context.status_db_ = raredisease_delivery_report_store_context
    delivery_report_context.meta_apis["analysis_api"] = RarediseaseAnalysisAPI(
        config=delivery_report_context
    )

    # Raredisease delivery report mocked methods
    mocker.patch.object(
        NfAnalysisAPI, "get_multiqc_json_path", return_value=raredisease_multiqc_json_metrics_path
    )
    mocker.patch.object(
        RarediseaseAnalysisAPI, "get_sample_coverage", return_value=coverage_metrics
    )
    mocker.patch.object(
        RarediseaseDeliveryReportAPI, "get_scout_variants_files", return_value=scout_variants_files
    )
    return delivery_report_context


@pytest.fixture
def rnafusion_delivery_report_context(
    coverage_metrics: CoverageMetrics,
    delivery_report_context: CGConfig,
    mocker: MockFixture,
    rnafusion_delivery_report_store_context: Store,
    rnafusion_multiqc_json_metrics_path: Path,
    scout_variants_files: ScoutVariantsFiles,
) -> CGConfig:
    """Rnafusion context for the delivery report generation."""

    # Rnafusion delivery report context
    delivery_report_context.status_db_ = rnafusion_delivery_report_store_context
    delivery_report_context.meta_apis["analysis_api"] = RnafusionAnalysisAPI(
        config=delivery_report_context
    )

    # Rnafusion delivery report mocked methods
    mocker.patch.object(
        NfAnalysisAPI, "get_multiqc_json_path", return_value=rnafusion_multiqc_json_metrics_path
    )
    mocker.patch.object(
        RnafusionDeliveryReportAPI, "get_scout_variants_files", return_value=scout_variants_files
    )
    return delivery_report_context


@pytest.fixture
def raredisease_delivery_report_click_context(
    raredisease_delivery_report_context: CGConfig,
) -> click.Context:
    """Raredisease click delivery report context fixture."""
    return click.Context(generate_delivery_report, obj=raredisease_delivery_report_context)



"""Tests for the read function pertaining to the Illumina Sequencing Run model."""

from datetime import datetime

import pytest

from cg.constants import SequencingRunDataAvailability, Workflow
from cg.exc import CgError
from cg.store.models import Case, IlluminaSequencingRun
from cg.store.store import Store


def test_get_latest_illumina_sequencing_run_for_nipt_case(
    re_sequenced_sample_illumina_data_store: Store,
    case_id_for_sample_on_multiple_flow_cells: str,
    flow_cells_with_the_same_sample: list[str],
    timestamp_yesterday: datetime,
):
    # GIVEN a store with a sample that has been sequenced twice
    first_run: IlluminaSequencingRun = (
        re_sequenced_sample_illumina_data_store.get_illumina_sequencing_run_by_device_internal_id(
            flow_cells_with_the_same_sample[0]
        )
    )
    first_run.sequencing_completed_at = timestamp_yesterday
    second_run: IlluminaSequencingRun = (
        re_sequenced_sample_illumina_data_store.get_illumina_sequencing_run_by_device_internal_id(
            flow_cells_with_the_same_sample[1]
        )
    )
    assert first_run
    assert second_run
    assert first_run.sequencing_completed_at < second_run.sequencing_completed_at

    # WHEN fetching the latest sequencing run for the sample
    latest_sequencing_run: IlluminaSequencingRun = (
        re_sequenced_sample_illumina_data_store.get_latest_illumina_sequencing_run_for_nipt_case(
            case_id_for_sample_on_multiple_flow_cells
        )
    )

    # THEN the latest sequencing run is returned
    assert latest_sequencing_run.device.internal_id == second_run.device.internal_id


def test_get_latest_illumina_sequencing_run_for_nipt_case_fail(
    re_sequenced_sample_illumina_data_store: Store,
    case_id_for_sample_on_multiple_flow_cells: str,
    flow_cells_with_the_same_sample: list[str],
    timestamp_yesterday: datetime,
):
    # GIVEN a store with a sample that has been sequenced twice but is not a NIPT case
    case: Case = re_sequenced_sample_illumina_data_store.get_case_by_internal_id(
        case_id_for_sample_on_multiple_flow_cells
    )
    case.data_analysis = Workflow.MIP_DNA

    # WHEN fetching the latest sequencing run for the sample

    # THEN an error is raised
    with pytest.raises(CgError):
        re_sequenced_sample_illumina_data_store.get_latest_illumina_sequencing_run_for_nipt_case(
            case_id_for_sample_on_multiple_flow_cells
        )


def test_get_illumina_sequencing_run_by_data_availability(
    store_with_illumina_sequencing_data: Store,
):
    # GIVEN a store with Illumina Sequencing Runs with data availability

    # WHEN filtering sequencing runs by data availability
    sequencing_runs: list[IlluminaSequencingRun] = (
        store_with_illumina_sequencing_data.get_illumina_sequencing_runs_by_data_availability(
            [SequencingRunDataAvailability.ON_DISK]
        )
    )

    # THEN the runs with the availability are returned
    for sequencing_run in sequencing_runs:
        assert sequencing_run.data_availability == SequencingRunDataAvailability.ON_DISK


def test_get_illumina_sequencing_run_by_data_availability_non_existetnt_status(
    store_with_illumina_sequencing_data: Store,
):
    # GIVEN a store with Illumina Sequencing Runs with data availability

    # WHEN filtering sequencing runs by data availability
    sequencing_runs: list[IlluminaSequencingRun] = (
        store_with_illumina_sequencing_data.get_illumina_sequencing_runs_by_data_availability(
            ["non_existent_status"]
        )
    )

    # THEN no runs are returned
    assert not sequencing_runs



from pathlib import Path

import pytest

from cg.constants import FileExtensions
from cg.utils.files import (
    get_directories_in_path,
    get_file_in_directory,
    get_file_with_pattern_from_list,
    get_files_in_directory_with_pattern,
    get_files_matching_pattern,
    get_project_root_dir,
    get_source_creation_time_stamp,
    remove_directory_and_contents,
    rename_file,
)


def test_get_project_root_dir():
    # WHEN getting the project root dir
    root_dir: Path = get_project_root_dir()

    # THEN return the dir path
    assert root_dir.name == "cg"


def test_get_file_in_directory(nested_directory_with_file: Path, some_file: str):
    """Test function to get a file in a directory and subdirectories."""
    # GIVEN a directory with subdirectories with a file

    # WHEN getting the file
    file_path: Path = get_file_in_directory(nested_directory_with_file, some_file)

    # THEN assert that the file is returned
    assert file_path.exists()


def test_get_file_with_pattern_from_list():
    """Test that a file is extracted from a list by pattern."""
    # GIVEN a list of files and a pattern found in one of the files
    files = [Path("file1.txt"), Path("file2.txt"), Path("file3.txt")]

    # WHEN getting a file by the pattern
    file = get_file_with_pattern_from_list(files, "file2")

    # THEN assert that the file is returned
    assert file == Path("file2.txt")


def test_get_file_with_pattern_from_list_no_file():
    """Test that a file is extracted from a list by pattern."""
    # GIVEN a list of files and a pattern not found in any of the files
    files = [Path("file1.txt"), Path("file2.txt"), Path("file3.txt")]

    # WHEN getting a file by the pattern

    # THEN a FileNotFoundError should be raised
    with pytest.raises(FileNotFoundError):
        get_file_with_pattern_from_list(files, "file4")


def test_get_files_in_directory_by_pattern(nested_directory_with_file: Path, some_file: str):
    """Test function to get files with a pattern in a directory and subdirectories."""
    # GIVEN a directory with subdirectories with a file

    # WHEN getting the file
    file_paths: list[Path] = get_files_in_directory_with_pattern(
        directory=nested_directory_with_file, pattern=some_file
    )

    # THEN assert that the file is returned
    for file_path in file_paths:
        assert file_path.exists()
        assert some_file in file_path.as_posix()


def test_get_files_matching_pattern(nested_directory_with_file: Path, some_file: str):
    # GIVEN a directory with a subdirectory containing a .txt file
    directory_with_file = Path(nested_directory_with_file, "sub_directory")

    # WHEN getting the file from the directory
    files = get_files_matching_pattern(
        directory=directory_with_file, pattern=f"*{FileExtensions.TXT}"
    )

    # THEN assert that the file is returned
    assert len(files) == 1


def test_get_files_matching_pattern_no_files(nested_directory_with_file: Path, some_file: str):
    # GIVEN a directory with a subdirectory containing a .txt file

    # WHEN getting the file from the directory
    files = get_files_matching_pattern(
        directory=nested_directory_with_file, pattern=f"*{FileExtensions.JSON}"
    )

    # THEN assert that the file is returned
    assert len(files) == 0


def test_rename_file(tmp_path: Path):
    # GIVEN a file path and a renamed file path
    file_path: Path = Path(tmp_path, "dummy_path")
    renamed_file_path: Path = Path(tmp_path, "dummy_renamed_path")

    # GIVEN that the file path exist
    file_path.touch()
    assert file_path.exists()

    # GIVEN that the renamed file path does not exist
    assert not renamed_file_path.exists()

    # WHEN renaming the file
    rename_file(file_path=file_path, renamed_file_path=renamed_file_path)

    # THEN the renamed file path should exist
    assert renamed_file_path.exists()

    # THEN the file path should not exist
    assert not file_path.exists()


def test_rename_file_exists(tmp_path: Path):
    # GIVEN a file path and a renamed file path
    file_path: Path = Path(tmp_path, "dummy_path")
    renamed_file_path: Path = Path(tmp_path, "dummy_renamed_path")

    # GIVEN that the file path exist
    file_path.touch()
    renamed_file_path.touch()
    assert renamed_file_path.exists()

    # WHEN renaming the file
    rename_file(file_path=file_path, renamed_file_path=renamed_file_path)

    # THEN the renamed file path should exist
    assert renamed_file_path.exists()


def test_rename_file_original_does_not_exist(tmp_path: Path):
    # GIVEN a file path and a renamed file path

    file_path: Path = Path(tmp_path, "dummy_path")
    renamed_file_path: Path = Path(tmp_path, "dummy_renamed_path")

    # GIVEN that the file path does not exist
    assert not file_path.exists()

    # WHEN renaming the file
    # THEN a FileNotFoundError should be raised
    with pytest.raises(FileNotFoundError):
        rename_file(file_path=file_path, renamed_file_path=renamed_file_path)


def test_get_creation_date_for_dir(tmp_path_factory):
    """Test to retrieve the creation date of a directory."""

    # GIVEN a directory that is created now
    directory_path: Path = tmp_path_factory.mktemp("some_dir")

    # WHEN retrieving the creation time stamp
    creation_time_stamp: float = get_source_creation_time_stamp(directory_path)

    # THEN the creation time stamp is returned
    assert isinstance(creation_time_stamp, float)


def test_get_creation_date_for_file(tmp_path_factory):
    """Test to retrieve the creation date of a directory."""

    # GIVEN a file that is created now
    directory_path: Path = tmp_path_factory.mktemp("some_dir")
    file_path: Path = Path(directory_path, "some_file")
    file_path.touch()

    # WHEN retrieving the creation time stamp
    creation_time_stamp: float = get_source_creation_time_stamp(file_path)

    # THEN the creation time stamp is returned
    assert isinstance(creation_time_stamp, float)


def test_get_all_directories_in_path(
    path_with_directories_and_a_file: Path, sub_dir_names: list[str], some_file: str
):
    """Test that get all directories in path only returns directories."""

    # GIVEN a path that contains directories and a file
    assert Path(path_with_directories_and_a_file, some_file).exists()

    # WHEN retrieving all directories in the path
    directories: list[Path] = get_directories_in_path(path=path_with_directories_and_a_file)

    # THEN all directories are returned
    for sub_dir_name in sub_dir_names:
        assert sub_dir_name in [directory.name for directory in directories]

    # THEN no files are returned
    assert some_file not in [directory.name for directory in directories]


def test_remove_directory_and_contents(path_with_directories_and_a_file: Path, some_file: str):
    """Test to remove a directory and all its contents."""

    # GIVEN a path to a directory
    assert path_with_directories_and_a_file.exists()

    # WHEN removing the directory
    remove_directory_and_contents(path_with_directories_and_a_file)

    # THEN the directory and its contents should no longer exist
    assert not path_with_directories_and_a_file.exists()
    assert not Path(path_with_directories_and_a_file, some_file).exists()



import logging
from pathlib import Path
from typing import Any

from pydantic import BaseModel, ConfigDict, EmailStr, Field
from typing_extensions import Literal

from cg.apps.coverage import ChanjoAPI
from cg.apps.crunchy import CrunchyAPI
from cg.apps.demultiplex.demultiplex_api import DemultiplexingAPI
from cg.apps.demultiplex.sample_sheet.api import IlluminaSampleSheetService
from cg.apps.gens import GensAPI
from cg.apps.gt import GenotypeAPI
from cg.apps.hermes.hermes_api import HermesApi
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.lims import LimsAPI
from cg.apps.loqus import LoqusdbAPI
from cg.apps.madeline.api import MadelineAPI
from cg.apps.mutacc_auto import MutaccAutoAPI
from cg.apps.scout.scoutapi import ScoutAPI
from cg.apps.tb import TrailblazerAPI
from cg.clients.arnold.api import ArnoldAPIClient
from cg.clients.chanjo2.client import Chanjo2APIClient
from cg.clients.janus.api import JanusAPIClient
from cg.constants.observations import LoqusdbInstance
from cg.constants.priority import SlurmQos
from cg.meta.delivery.delivery import DeliveryAPI
from cg.services.analysis_service.analysis_service import AnalysisService
from cg.services.decompression_service.decompressor import Decompressor
from cg.services.deliver_files.deliver_files_service.deliver_files_service_factory import (
    DeliveryServiceFactory,
)
from cg.services.deliver_files.rsync.models import RsyncDeliveryConfig
from cg.services.deliver_files.rsync.service import DeliveryRsyncService
from cg.services.fastq_concatenation_service.fastq_concatenation_service import (
    FastqConcatenationService,
)
from cg.services.pdc_service.pdc_service import PdcService
from cg.services.run_devices.pacbio.data_storage_service.pacbio_store_service import (
    PacBioStoreService,
)
from cg.services.run_devices.pacbio.data_transfer_service.data_transfer_service import (
    PacBioDataTransferService,
)
from cg.services.run_devices.pacbio.housekeeper_service.pacbio_houskeeper_service import (
    PacBioHousekeeperService,
)
from cg.services.run_devices.pacbio.metrics_parser.metrics_parser import (
    PacBioMetricsParser,
)
from cg.services.run_devices.pacbio.post_processing_service import (
    PacBioPostProcessingService,
)
from cg.services.run_devices.pacbio.run_data_generator.pacbio_run_data_generator import (
    PacBioRunDataGenerator,
)
from cg.services.run_devices.pacbio.run_file_manager.run_file_manager import (
    PacBioRunFileManager,
)
from cg.services.run_devices.pacbio.run_validator.pacbio_run_validator import (
    PacBioRunValidator,
)
from cg.services.run_devices.run_names.pacbio import PacbioRunNamesService
from cg.services.sequencing_qc_service.sequencing_qc_service import SequencingQCService
from cg.services.slurm_service.slurm_cli_service import SlurmCLIService
from cg.services.slurm_service.slurm_service import SlurmService
from cg.services.slurm_upload_service.slurm_upload_config import SlurmUploadConfig
from cg.services.slurm_upload_service.slurm_upload_service import SlurmUploadService
from cg.services.validate_file_transfer_service.validate_file_transfer_service import (
    ValidateFileTransferService,
)
from cg.store.database import initialize_database
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class Sequencers(BaseModel):
    hiseqx: str
    hiseqga: str
    novaseq: str


class ArnoldConfig(BaseModel):
    api_url: str


class SlurmConfig(BaseModel):
    account: str
    hours: int | None = None
    mail_user: EmailStr
    memory: int | None = None
    number_tasks: int | None = None
    conda_env: str | None = None
    qos: SlurmQos = SlurmQos.LOW


class Encryption(BaseModel):
    encryption_dir: str
    binary_path: str


class PDCArchivingDirectory(BaseModel):
    current: str
    nas: str
    pre_nas: str


class DataInput(BaseModel):
    input_dir_path: str


class IlluminaBackupConfig(BaseModel):
    pdc_archiving_directory: PDCArchivingDirectory
    slurm_flow_cell_encryption: SlurmConfig


class HousekeeperConfig(BaseModel):
    database: str
    root: str


class DemultiplexConfig(BaseModel):
    slurm: SlurmConfig


class DownsampleConfig(BaseModel):
    downsample_dir: str
    downsample_script: str
    account: str


class ClientConfig(BaseModel):
    host: str


class TrailblazerConfig(BaseModel):
    service_account: str
    service_account_auth_file: str
    host: str


class StatinaConfig(BaseModel):
    host: str | None = None
    user: str
    key: str
    api_url: str
    upload_path: str
    auth_path: str


class CommonAppConfig(BaseModel):
    binary_path: str | None = None
    config_path: str | None = None


class FluffyUploadConfig(BaseModel):
    user: str
    password: str
    host: str
    remote_path: str
    port: int


class FluffyConfig(CommonAppConfig):
    root_dir: str
    sftp: FluffyUploadConfig


class LimsConfig(BaseModel):
    host: str
    username: str
    password: str


class CrunchyConfig(BaseModel):
    conda_binary: str | None = None
    cram_reference: str
    slurm: SlurmConfig


class MutaccAutoConfig(CommonAppConfig):
    padding: int = 300


class BalsamicConfig(CommonAppConfig):
    balsamic_cache: str
    bed_path: str
    binary_path: str
    cadd_path: str
    conda_binary: str
    conda_env: str
    genome_interval_path: str
    gens_coverage_female_path: str
    gens_coverage_male_path: str
    gnomad_af5_path: str
    loqusdb_path: str
    pon_path: str
    root: str
    sentieon_licence_path: str
    slurm: SlurmConfig
    swegen_path: str


class MutantConfig(BaseModel):
    binary_path: str
    conda_binary: str | None = None
    conda_env: str
    root: str


class MipConfig(BaseModel):
    conda_binary: str | None = None
    conda_env: str
    mip_config: str
    workflow: str
    root: str
    script: str


class RarediseaseConfig(CommonAppConfig):
    binary_path: str | None = None
    compute_env: str
    conda_binary: str | None = None
    conda_env: str
    config_platform: str
    config_params: str
    config_resources: str
    launch_directory: str
    workflow_path: str
    profile: str
    references: str
    revision: str
    root: str
    slurm: SlurmConfig
    tower_workflow: str


class TomteConfig(CommonAppConfig):
    binary_path: str | None = None
    compute_env: str
    conda_binary: str | None = None
    conda_env: str
    config_platform: str
    config_params: str
    config_resources: str
    workflow_path: str
    profile: str
    references: str
    revision: str
    root: str
    slurm: SlurmConfig
    tower_workflow: str


class RnafusionConfig(CommonAppConfig):
    binary_path: str
    compute_env: str
    conda_binary: str | None = None
    conda_env: str
    config_platform: str
    config_params: str
    config_resources: str
    launch_directory: str
    profile: str
    references: str
    revision: str
    root: str
    slurm: SlurmConfig
    tower_workflow: str
    workflow_path: str


class TaxprofilerConfig(CommonAppConfig):
    binary_path: str
    conda_binary: str | None = None
    conda_env: str
    compute_env: str
    databases: str
    hostremoval_reference: str
    workflow_path: str
    profile: str
    revision: str
    root: str
    slurm: SlurmConfig
    tower_workflow: str


class MicrosaltConfig(BaseModel):
    binary_path: str
    conda_binary: str | None = None
    conda_env: str
    queries_path: str
    root: str


class GisaidConfig(CommonAppConfig):
    submitter: str
    log_dir: str
    logwatch_email: EmailStr
    upload_password: str
    upload_cid: str


class DataDeliveryConfig(BaseModel):
    account: str
    base_path: str
    covid_destination_path: str
    covid_report_path: str
    destination_path: str
    mail_user: str


class EmailBaseSettings(BaseModel):
    sender_email: EmailStr
    smtp_server: str


class FOHMConfig(BaseModel):
    host: str
    port: int
    key: str
    username: str
    valid_uploader: str
    email_sender: str
    email_recipient: str
    email_host: str


class ExternalConfig(BaseModel):
    hasta: str
    caesar: str


class DataFlowConfig(BaseModel):
    database_name: str
    user: str
    password: str
    url: str
    local_storage: str
    archive_repository: str


class PacbioConfig(BaseModel):
    data_dir: str
    systemd_trigger_dir: str


class OxfordNanoporeConfig(BaseModel):
    data_dir: str
    systemd_trigger_dir: str


class IlluminaConfig(BaseModel):
    sequencing_runs_dir: str
    demultiplexed_runs_dir: str


class RunInstruments(BaseModel):
    pacbio: PacbioConfig
    nanopore: OxfordNanoporeConfig
    illumina: IlluminaConfig


class RunNamesServices(BaseModel):
    pacbio: PacbioRunNamesService
    model_config = ConfigDict(arbitrary_types_allowed=True)


class PostProcessingServices(BaseModel):
    pacbio: PacBioPostProcessingService
    model_config = ConfigDict(arbitrary_types_allowed=True)


class CGConfig(BaseModel):
    data_input: DataInput | None = None
    database: str
    delivery_path: str
    downsample: DownsampleConfig
    email_base_settings: EmailBaseSettings
    environment: Literal["production", "stage"] = "stage"
    madeline_exe: str
    max_flowcells: int | None = None
    nanopore_data_directory: str
    run_instruments: RunInstruments
    sentieon_licence_server: str
    tower_binary_path: str

    # Base APIs that always should exist
    housekeeper: HousekeeperConfig
    housekeeper_api_: HousekeeperAPI = None
    status_db_: Store | None = None

    # App APIs that can be instantiated in CGConfig
    arnold: ArnoldConfig | None = None
    arnold_api_: ArnoldAPIClient | None = None
    illumina_backup_service: IlluminaBackupConfig | None = None
    chanjo: CommonAppConfig = None
    chanjo_api_: ChanjoAPI = None
    chanjo2: ClientConfig | None = None
    chanjo2_api_: Chanjo2APIClient | None = None
    crunchy: CrunchyConfig = None
    crunchy_api_: CrunchyAPI = None
    data_delivery: DataDeliveryConfig = Field(None, alias="data-delivery")
    data_flow: DataFlowConfig | None = None
    delivery_api_: DeliveryAPI | None = None
    delivery_rsync_service_: DeliveryRsyncService | None = None
    delivery_service_factory_: DeliveryServiceFactory | None = None
    demultiplex: DemultiplexConfig = None
    demultiplex_api_: DemultiplexingAPI = None
    encryption: Encryption | None = None
    external: ExternalConfig = None
    genotype: CommonAppConfig = None
    genotype_api_: GenotypeAPI = None
    gens: CommonAppConfig = None
    gens_api_: GensAPI = None
    hermes: CommonAppConfig = None
    hermes_api_: HermesApi = None
    janus: ClientConfig | None = None
    janus_api_: JanusAPIClient | None = None
    lims: LimsConfig = None
    lims_api_: LimsAPI = None
    loqusdb: CommonAppConfig = Field(None, alias=LoqusdbInstance.WGS.value)
    loqusdb_api_: LoqusdbAPI = None
    loqusdb_somatic: CommonAppConfig = Field(None, alias=LoqusdbInstance.SOMATIC.value)
    loqusdb_tumor: CommonAppConfig = Field(None, alias=LoqusdbInstance.TUMOR.value)
    loqusdb_wes: CommonAppConfig = Field(None, alias=LoqusdbInstance.WES.value)
    madeline_api_: MadelineAPI = None
    mutacc_auto: MutaccAutoConfig = Field(None, alias="mutacc-auto")
    mutacc_auto_api_: MutaccAutoAPI = None
    pdc: CommonAppConfig | None = None
    pdc_service_: PdcService | None = None
    post_processing_services_: PostProcessingServices | None = None
    pigz: CommonAppConfig | None = None
    run_names_services_: RunNamesServices | None = None
    sample_sheet_api_: IlluminaSampleSheetService | None = None
    scout: CommonAppConfig = None
    scout_api_: ScoutAPI = None
    tar: CommonAppConfig | None = None
    trailblazer: TrailblazerConfig = None
    trailblazer_api_: TrailblazerAPI = None

    # Meta APIs that will use the apps from CGConfig
    balsamic: BalsamicConfig | None = None
    fluffy: FluffyConfig | None = None
    fohm: FOHMConfig | None = None
    gisaid: GisaidConfig | None = None
    microsalt: MicrosaltConfig | None = None
    mip_rd_dna: MipConfig | None = Field(None, alias="mip-rd-dna")
    mip_rd_rna: MipConfig | None = Field(None, alias="mip-rd-rna")
    mutant: MutantConfig | None = None
    raredisease: RarediseaseConfig | None = None
    rnafusion: RnafusionConfig | None = None
    statina: StatinaConfig | None = None
    taxprofiler: TaxprofilerConfig | None = None
    tomte: TomteConfig | None = None

    # These are meta APIs that gets instantiated in the code
    meta_apis: dict = {}

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @property
    def arnold_api(self) -> ArnoldAPIClient:
        api = self.__dict__.get("arnold_api_")
        if api is None:
            LOG.debug("Instantiating arnold api")
            api = ArnoldAPIClient(config=self.dict())
            self.arnold_api_ = api
        return api

    @property
    def chanjo_api(self) -> ChanjoAPI:
        api = self.__dict__.get("chanjo_api_")
        if api is None:
            LOG.debug("Instantiating chanjo api")
            api = ChanjoAPI(config=self.dict())
            self.chanjo_api_ = api
        return api

    @property
    def chanjo2_api(self) -> Chanjo2APIClient:
        chanjo2_api = self.__dict__.get("chanjo2_api_")
        if chanjo2_api is None:
            LOG.debug("Instantiating Chanjo2 API")
            config: dict[str, Any] = self.dict()
            chanjo2_api = Chanjo2APIClient(base_url=config["chanjo2"]["host"])
            self.chanjo2_api_ = chanjo2_api
        return chanjo2_api

    @property
    def crunchy_api(self) -> CrunchyAPI:
        api = self.__dict__.get("crunchy_api_")
        if api is None:
            LOG.debug("Instantiating crunchy api")
            api = CrunchyAPI(config=self.dict())
            self.crunchy_api_ = api
        return api

    @property
    def demultiplex_api(self) -> DemultiplexingAPI:
        demultiplex_api = self.__dict__.get("demultiplex_api_")
        if demultiplex_api is None:
            LOG.debug("Instantiating demultiplexing api")
            demultiplex_api = DemultiplexingAPI(
                config=self.dict(), housekeeper_api=self.housekeeper_api
            )
            self.demultiplex_api_ = demultiplex_api
        return demultiplex_api

    @property
    def genotype_api(self) -> GenotypeAPI:
        api = self.__dict__.get("genotype_api_")
        if api is None:
            LOG.debug("Instantiating genotype api")
            api = GenotypeAPI(config=self.dict())
            self.genotype_api_ = api
        return api

    @property
    def gens_api(self) -> GensAPI:
        """Returns Gens API after making sure it has been instantiated."""
        api = self.__dict__.get("gens_api_")
        if api is None:
            LOG.debug("Instantiating gens api")
            api = GensAPI(config=self.dict())
            self.gens_api_ = api
        return api

    @property
    def hermes_api(self) -> HermesApi:
        hermes_api = self.__dict__.get("hermes_api_")
        if hermes_api is None:
            LOG.debug("Instantiating hermes api")
            hermes_api = HermesApi(config=self.dict())
            self.hermes_api_ = hermes_api
        return hermes_api

    @property
    def housekeeper_api(self) -> HousekeeperAPI:
        housekeeper_api = self.__dict__.get("housekeeper_api_")
        if housekeeper_api is None:
            LOG.debug("Instantiating housekeeper api")
            housekeeper_api = HousekeeperAPI(config=self.dict())
            self.housekeeper_api_ = housekeeper_api
        return housekeeper_api

    @property
    def janus_api(self) -> JanusAPIClient:
        janus_api = self.__dict__.get("janus_api_")
        if janus_api is None:
            LOG.debug("Instantiating janus api")
            janus_api = JanusAPIClient(config=self.dict())
            self.janus_api_ = janus_api
        return janus_api

    @property
    def lims_api(self) -> LimsAPI:
        api = self.__dict__.get("lims_api_")
        if api is None:
            LOG.debug("Instantiating lims api")
            api = LimsAPI(config=self.dict())
            self.lims_api_ = api
        return api

    @property
    def loqusdb_api(self) -> LoqusdbAPI:
        api = self.__dict__.get("loqusdb_api_")
        if api is None:
            LOG.debug("Instantiating loqusdb api")
            api: LoqusdbAPI = LoqusdbAPI(
                binary_path=self.loqusdb.binary_path, config_path=self.loqusdb.config_path
            )
            self.loqusdb_api_ = api
        return api

    @property
    def madeline_api(self) -> MadelineAPI:
        api = self.__dict__.get("madeline_api_")
        if api is None:
            LOG.debug("Instantiating madeline api")
            api = MadelineAPI(config=self.dict())
            self.madeline_api_ = api
        return api

    @property
    def mutacc_auto_api(self) -> MutaccAutoAPI:
        api = self.__dict__.get("mutacc_auto_api_")
        if api is None:
            LOG.debug("Instantiating mutacc_auto api")
            api = MutaccAutoAPI(config=self.dict())
            self.mutacc_auto_api_ = api
        return api

    @property
    def post_processing_services(self) -> PostProcessingServices:
        services = self.__dict__.get("post_processing_services_")
        if services is None:
            LOG.debug("Instantiating post-processing services")
            services = PostProcessingServices(
                pacbio=self.get_pacbio_post_processing_service(),
            )
            self.post_processing_services_ = services
        return services

    def get_pacbio_post_processing_service(self) -> PacBioPostProcessingService:
        LOG.debug("Instantiating PacBio post-processing service")
        run_data_generator = PacBioRunDataGenerator()
        file_manager = PacBioRunFileManager()
        run_validator = PacBioRunValidator(
            file_manager=file_manager,
            decompressor=Decompressor(),
            file_transfer_validator=ValidateFileTransferService(),
        )
        metrics_parser = PacBioMetricsParser(file_manager=file_manager)
        transfer_service = PacBioDataTransferService(metrics_service=metrics_parser)
        store_service = PacBioStoreService(
            store=self.status_db, data_transfer_service=transfer_service
        )
        hk_service = PacBioHousekeeperService(
            hk_api=self.housekeeper_api,
            file_manager=file_manager,
            metrics_parser=metrics_parser,
        )
        return PacBioPostProcessingService(
            run_validator=run_validator,
            run_data_generator=run_data_generator,
            hk_service=hk_service,
            store_service=store_service,
            sequencing_dir=self.run_instruments.pacbio.data_dir,
        )

    @property
    def pdc_service(self) -> PdcService:
        service = self.__dict__.get("pdc_service_")
        if service is None:
            LOG.debug("Instantiating PDC service")
            service = PdcService(binary_path=self.pdc.binary_path)
            self.pdc_service_ = service
        return service

    @property
    def run_names_services(self) -> RunNamesServices:
        services = self.run_names_services_
        if services is None:
            LOG.debug("Instantiating run directory names services")
            services = RunNamesServices(
                pacbio=PacbioRunNamesService(self.run_instruments.pacbio.data_dir)
            )
            self.run_names_services_ = services
        return services

    @property
    def sample_sheet_api(self) -> IlluminaSampleSheetService:
        sample_sheet_api = self.__dict__.get("sample_sheet_api_")
        if sample_sheet_api is None:
            LOG.debug("Instantiating sample sheet API")
            sample_sheet_api = IlluminaSampleSheetService(
                flow_cell_dir=self.run_instruments.illumina.sequencing_runs_dir,
                hk_api=self.housekeeper_api,
                lims_api=self.lims_api,
            )
            self.sample_sheet_api_ = sample_sheet_api
        return sample_sheet_api

    @property
    def slurm_service(self) -> SlurmService:
        return SlurmCLIService()

    @property
    def slurm_upload_service(self) -> SlurmUploadService:
        slurm_upload_config = SlurmUploadConfig(
            email=self.data_delivery.mail_user,
            account=self.data_delivery.account,
            log_dir=self.data_delivery.base_path,
        )
        return SlurmUploadService(
            slurm_service=self.slurm_service,
            trailblazer_api=self.trailblazer_api,
            config=slurm_upload_config,
        )

    @property
    def analysis_service(self) -> AnalysisService:
        return AnalysisService(analysis_client=self.trailblazer_api, status_db=self.status_db)

    @property
    def scout_api(self) -> ScoutAPI:
        api = self.__dict__.get("scout_api_")
        if api is None:
            LOG.debug("Instantiating scout api")
            api = ScoutAPI(config=self.dict(), slurm_upload_service=self.slurm_upload_service)
            self.scout_api_ = api
        return api

    @property
    def status_db(self) -> Store:
        status_db = self.__dict__.get("status_db_")
        if status_db is None:
            LOG.debug("Instantiating status db")
            initialize_database(self.database)
            status_db = Store()
            self.status_db_ = status_db
        return status_db

    @property
    def trailblazer_api(self) -> TrailblazerAPI:
        api = self.__dict__.get("trailblazer_api_")
        if api is None:
            LOG.debug("Instantiating trailblazer api")
            api = TrailblazerAPI(config=self.dict())
            self.trailblazer_api_ = api
        return api

    @property
    def fastq_concatenation_service(self) -> FastqConcatenationService:
        return FastqConcatenationService()

    @property
    def delivery_api(self) -> DeliveryAPI:
        api = self.__dict__.get("delivery_api_")
        if api is None:
            LOG.debug("Instantiating delivery api")
            api = DeliveryAPI(
                delivery_path=Path(self.delivery_path),
                fastq_concatenation_service=self.fastq_concatenation_service,
                housekeeper_api=self.housekeeper_api,
                store=self.status_db,
            )
            self.delivery_api_ = api
        return api

    @property
    def sequencing_qc_service(self) -> SequencingQCService:
        return SequencingQCService(self.status_db)

    @property
    def delivery_rsync_service(self) -> DeliveryRsyncService:
        service = self.delivery_rsync_service_
        if service is None:
            LOG.debug("Instantiating delivery rsync service")
            rsync_config = RsyncDeliveryConfig(**self.data_delivery.dict())
            service = DeliveryRsyncService(
                delivery_path=self.delivery_path,
                rsync_config=rsync_config,
                status_db=self.status_db,
            )
            self.delivery_rsync_service_ = service
        return service

    @property
    def delivery_service_factory(self) -> DeliveryServiceFactory:
        factory = self.delivery_service_factory_
        if not factory:
            LOG.debug("Instantiating delivery service factory")
            factory = DeliveryServiceFactory(
                store=self.status_db,
                hk_api=self.housekeeper_api,
                tb_service=self.trailblazer_api,
                rsync_service=self.delivery_rsync_service,
                analysis_service=self.analysis_service,
            )
            self.delivery_service_factory_ = factory
        return factory



from pathlib import Path

from click.testing import CliRunner
from cg.cli.workflow.raredisease.base import panel, managed_variants
from cg.constants.scout import ScoutExportFileName
from cg.io.txt import read_txt
from cg.meta.workflow.raredisease import RarediseaseAnalysisAPI
from cg.models.cg_config import CGConfig


def test_panel_dry_run(
    raredisease_case_id: str,
    cli_runner: CliRunner,
    raredisease_context: CGConfig,
    scout_panel_output: str,
    mocker,
):
    # GIVEN a case

    # GIVEN that, the Scout command writes the panel to stdout
    mocker.patch.object(RarediseaseAnalysisAPI, "get_gene_panel", return_value=scout_panel_output)

    result = cli_runner.invoke(panel, [raredisease_case_id, "--dry-run"], obj=raredisease_context)
    # THEN the output should contain the output from Scout
    actual_output = "".join(result.stdout.strip().split())
    expected_output = "".join(scout_panel_output.strip().split())

    assert actual_output == expected_output


def test_panel_file_is_written(
    raredisease_case_id: str,
    cli_runner: CliRunner,
    raredisease_context: CGConfig,
    scout_panel_output: str,
    mocker,
):
    # GIVEN an analysis API
    analysis_api: RarediseaseAnalysisAPI = raredisease_context.meta_apis["analysis_api"]

    # GIVEN a case

    # GIVEN that, the Scout command writes the panel to stdout
    mocker.patch.object(RarediseaseAnalysisAPI, "get_gene_panel", return_value=scout_panel_output)

    cli_runner.invoke(panel, [raredisease_case_id], obj=raredisease_context)

    panel_file = Path(analysis_api.root, raredisease_case_id, ScoutExportFileName.PANELS)

    # THEN the file should exist
    assert panel_file.exists()

    # THEN the file should contain the output from Scout
    file_content: str = read_txt(file_path=panel_file, read_to_string=True)
    assert "".join(file_content.strip().split()) == "".join(scout_panel_output.strip().split())


def test_managed_variants_is_written(
    raredisease_case_id: str,
    cli_runner: CliRunner,
    raredisease_context: CGConfig,
    scout_export_manged_variants_output: str,
    mocker,
):
    # GIVEN an analysis API
    analysis_api: RarediseaseAnalysisAPI = raredisease_context.meta_apis["analysis_api"]

    # GIVEN a case

    # GIVEN a mocked scout export of the managed variants
    mocker.patch.object(
        RarediseaseAnalysisAPI,
        "get_managed_variants",
        return_value=scout_export_manged_variants_output,
    )
    managed_variants_file = Path(
        analysis_api.root, raredisease_case_id, ScoutExportFileName.MANAGED_VARIANTS
    )

    # WHEN creating a managed variants file
    cli_runner.invoke(managed_variants, [raredisease_case_id], obj=raredisease_context)

    # THEN the file should exist
    assert managed_variants_file.exists()

    # THEN the file should contain the output from Scout
    file_content: str = read_txt(file_path=managed_variants_file, read_to_string=True)
    assert "".join(file_content.strip().split()) == "".join(
        scout_export_manged_variants_output.strip().split()
    )


def test_managed_variants_dry_run(
    raredisease_case_id: str,
    cli_runner: CliRunner,
    raredisease_context: CGConfig,
    scout_export_manged_variants_output: str,
    mocker,
):
    # GIVEN a case

    # GIVEN a mocked scout export of the managed variants
    mocker.patch.object(
        RarediseaseAnalysisAPI,
        "get_managed_variants",
        return_value=scout_export_manged_variants_output,
    )
    result = cli_runner.invoke(
        managed_variants, [raredisease_case_id, "--dry-run"], obj=raredisease_context
    )

    # THEN the result should contain the output from Scout
    assert "".join(result.stdout.strip().split()) == "".join(
        scout_export_manged_variants_output.strip().split()
    )



from cg.server.dto.samples.collaborator_samples_request import (
    CollaboratorSamplesRequest,
)
from cg.server.dto.samples.samples_response import SamplesResponse
from cg.services.sample_service.utils import (
    create_samples_response,
    get_cancel_comment,
    get_confirmation_message,
)
from cg.store.models import Sample
from cg.store.store import Store


class SampleService:
    def __init__(self, store: Store):
        self.store = store

    def cancel_sample(self, sample_id: int, user_email: str) -> None:
        self.store.mark_sample_as_cancelled(sample_id)
        self.store.decouple_sample_from_cases(sample_id)
        self._add_cancel_comment(sample_id=sample_id, user_email=user_email)

    def _add_cancel_comment(self, sample_id: int, user_email: str) -> None:
        if user := self.store.get_user_by_email(user_email):
            comment = get_cancel_comment(user.name)
            self.store.update_sample_comment(sample_id=sample_id, comment=comment)

    def cancel_samples(self, sample_ids: list[int], user_email: str) -> str:
        case_ids = self.store.get_case_ids_for_samples(sample_ids)

        for sample_id in sample_ids:
            self.cancel_sample(sample_id=sample_id, user_email=user_email)

        self.store.delete_cases_without_samples(case_ids)
        remaining_cases = self.store.filter_cases_with_samples(case_ids)
        return get_confirmation_message(sample_ids=sample_ids, case_ids=remaining_cases)

    def get_collaborator_samples(self, request: CollaboratorSamplesRequest) -> SamplesResponse:
        samples: list[Sample] = self.store.get_collaborator_samples(request)
        return create_samples_response(samples)



import logging

from cg.constants import (
    BALSAMIC_ANALYSIS_TYPE,
    BALSAMIC_REPORT_ACCREDITED_PANELS,
    REQUIRED_APPLICATION_FIELDS,
    REQUIRED_CASE_FIELDS,
    REQUIRED_CUSTOMER_FIELDS,
    REQUIRED_DATA_ANALYSIS_BALSAMIC_FIELDS,
    REQUIRED_DATA_ANALYSIS_FIELDS,
    REQUIRED_REPORT_FIELDS,
    REQUIRED_SAMPLE_BALSAMIC_FIELDS,
    REQUIRED_SAMPLE_METADATA_BALSAMIC_TARGETED_FIELDS,
    REQUIRED_SAMPLE_METADATA_BALSAMIC_TN_WGS_FIELDS,
    REQUIRED_SAMPLE_METADATA_BALSAMIC_TO_WGS_FIELDS,
    REQUIRED_SAMPLE_METHODS_FIELDS,
    REQUIRED_SAMPLE_TIMESTAMP_FIELDS,
    Workflow,
)
from cg.constants.constants import AnalysisType
from cg.constants.scout import ScoutUploadKey
from cg.meta.delivery_report.data_validators import get_million_read_pairs
from cg.meta.delivery_report.delivery_report_api import DeliveryReportAPI
from cg.meta.workflow.balsamic import BalsamicAnalysisAPI
from cg.models.balsamic.analysis import BalsamicAnalysis
from cg.models.balsamic.metrics import BalsamicTargetedQCMetrics, BalsamicWGSQCMetrics
from cg.models.delivery_report.metadata import (
    BalsamicTargetedSampleMetadataModel,
    BalsamicWGSSampleMetadataModel,
)
from cg.models.delivery_report.report import CaseModel, ReportRequiredFields, ScoutVariantsFiles
from cg.models.delivery_report.sample import SampleModel
from cg.store.models import Bed, BedVersion, Case, Sample

LOG = logging.getLogger(__name__)


class BalsamicDeliveryReportAPI(DeliveryReportAPI):
    """API to create Balsamic delivery reports."""

    def __init__(self, analysis_api: BalsamicAnalysisAPI):
        super().__init__(analysis_api=analysis_api)

    def get_sample_metadata(
        self, case: Case, sample: Sample, analysis_metadata: BalsamicAnalysis
    ) -> BalsamicTargetedSampleMetadataModel | BalsamicWGSSampleMetadataModel:
        """Return sample metadata to include in the report."""
        sample_metrics: BalsamicTargetedQCMetrics | BalsamicWGSQCMetrics = (
            analysis_metadata.sample_metrics[sample.internal_id]
        )
        million_read_pairs: float = get_million_read_pairs(sample.reads)
        passed_initial_qc: bool | None = self.lims_api.has_sample_passed_initial_qc(
            sample.internal_id
        )
        if AnalysisType.WHOLE_GENOME_SEQUENCING in self.analysis_api.get_data_analysis_type(
            case.internal_id
        ):
            return self.get_wgs_metadata(
                million_read_pairs=million_read_pairs,
                passed_initial_qc=passed_initial_qc,
                sample_metrics=sample_metrics,
            )
        return self.get_panel_metadata(
            analysis_metadata=analysis_metadata,
            million_read_pairs=million_read_pairs,
            passed_initial_qc=passed_initial_qc,
            sample_metrics=sample_metrics,
        )

    def get_panel_metadata(
        self,
        million_read_pairs: float,
        passed_initial_qc: bool | None,
        sample_metrics: BalsamicTargetedQCMetrics,
        analysis_metadata: BalsamicAnalysis,
    ) -> BalsamicTargetedSampleMetadataModel:
        """Return report metadata for Balsamic TARGETED_GENOME_SEQUENCING analysis."""
        bed_version: BedVersion = self.status_db.get_bed_version_by_file_name(
            analysis_metadata.balsamic_config.panel.capture_kit
        )
        bed: Bed = self.status_db.get_bed_by_entry_id(bed_version.bed_id) if bed_version else None
        return BalsamicTargetedSampleMetadataModel(
            bait_set=bed.name if bed else None,
            bait_set_version=analysis_metadata.balsamic_config.panel.capture_kit_version,
            duplicates=sample_metrics.percent_duplication if sample_metrics else None,
            fold_80=sample_metrics.fold_80_base_penalty if sample_metrics else None,
            gc_dropout=sample_metrics.gc_dropout if sample_metrics else None,
            initial_qc=passed_initial_qc,
            mean_insert_size=sample_metrics.mean_insert_size if sample_metrics else None,
            median_target_coverage=(
                sample_metrics.median_target_coverage if sample_metrics else None
            ),
            million_read_pairs=million_read_pairs,
            pct_250x=sample_metrics.pct_target_bases_250x if sample_metrics else None,
            pct_500x=sample_metrics.pct_target_bases_500x if sample_metrics else None,
        )

    @staticmethod
    def get_wgs_metadata(
        million_read_pairs: float,
        passed_initial_qc: bool | None,
        sample_metrics: BalsamicWGSQCMetrics,
    ) -> BalsamicWGSSampleMetadataModel:
        """Return report metadata for Balsamic WHOLE_GENOME_SEQUENCING analysis."""
        return BalsamicWGSSampleMetadataModel(
            duplicates=sample_metrics.percent_duplication if sample_metrics else None,
            fold_80=sample_metrics.fold_80_base_penalty if sample_metrics else None,
            initial_qc=passed_initial_qc,
            mean_insert_size=sample_metrics.mean_insert_size if sample_metrics else None,
            median_coverage=sample_metrics.median_coverage if sample_metrics else None,
            million_read_pairs=million_read_pairs,
            pct_15x=sample_metrics.pct_15x if sample_metrics else None,
            pct_60x=sample_metrics.pct_60x if sample_metrics else None,
            pct_reads_improper_pairs=(
                sample_metrics.pct_pf_reads_improper_pairs if sample_metrics else None
            ),
        )

    def is_report_accredited(
        self, samples: list[SampleModel], analysis_metadata: BalsamicAnalysis
    ) -> bool:
        """Return whether the Balsamic delivery report is accredited."""
        if analysis_metadata.balsamic_config.analysis.sequencing_type == "targeted" and next(
            (
                panel
                for panel in BALSAMIC_REPORT_ACCREDITED_PANELS
                if panel in str(analysis_metadata.balsamic_config.panel.capture_kit)
            ),
            None,
        ):
            return True
        return False

    def get_scout_variants_files(self, case_id: str) -> ScoutVariantsFiles:
        """Return files that will be uploaded to Scout."""
        return ScoutVariantsFiles(
            snv_vcf=self.get_scout_uploaded_file_from_hk(
                case_id=case_id, scout_key=ScoutUploadKey.SNV_VCF
            ),
            sv_vcf=self.get_scout_uploaded_file_from_hk(
                case_id=case_id, scout_key=ScoutUploadKey.SV_VCF
            ),
        )

    def get_required_fields(self, case: CaseModel) -> dict:
        """Return a dictionary with the delivery report required fields for Balsamic."""
        analysis_type: str = case.data_analysis.type
        required_data_analysis_fields: list[str] = (
            REQUIRED_DATA_ANALYSIS_FIELDS
            if self.analysis_api.workflow == Workflow.BALSAMIC_QC
            else REQUIRED_DATA_ANALYSIS_BALSAMIC_FIELDS
        )
        required_sample_metadata_fields: list[str] = []
        if BALSAMIC_ANALYSIS_TYPE["tumor_wgs"] in analysis_type:
            required_sample_metadata_fields: list[str] = (
                REQUIRED_SAMPLE_METADATA_BALSAMIC_TO_WGS_FIELDS
            )
        elif BALSAMIC_ANALYSIS_TYPE["tumor_normal_wgs"] in analysis_type:
            required_sample_metadata_fields: list[str] = (
                REQUIRED_SAMPLE_METADATA_BALSAMIC_TN_WGS_FIELDS
            )
        elif (
            BALSAMIC_ANALYSIS_TYPE["tumor_panel"] in analysis_type
            or BALSAMIC_ANALYSIS_TYPE["tumor_normal_panel"] in analysis_type
        ):
            required_sample_metadata_fields: list[str] = (
                REQUIRED_SAMPLE_METADATA_BALSAMIC_TARGETED_FIELDS
            )

        report_required_fields = ReportRequiredFields(
            applications=self.get_application_required_fields(
                case=case, required_fields=REQUIRED_APPLICATION_FIELDS
            ),
            case=REQUIRED_CASE_FIELDS,
            customer=REQUIRED_CUSTOMER_FIELDS,
            data_analysis=required_data_analysis_fields,
            metadata=self.get_sample_required_fields(
                case=case, required_fields=required_sample_metadata_fields
            ),
            methods=self.get_sample_required_fields(
                case=case, required_fields=REQUIRED_SAMPLE_METHODS_FIELDS
            ),
            report=REQUIRED_REPORT_FIELDS,
            samples=self.get_sample_required_fields(
                case=case, required_fields=REQUIRED_SAMPLE_BALSAMIC_FIELDS
            ),
            timestamps=self.get_timestamp_required_fields(
                case=case, required_fields=REQUIRED_SAMPLE_TIMESTAMP_FIELDS
            ),
        )
        return report_required_fields.model_dump()



from cg.constants import Workflow
from cg.constants.housekeeper_tags import AlignmentFileTag
from cg.services.deliver_files.tag_fetcher.abstract import (
    FetchDeliveryFileTagsService,
)
from cg.services.deliver_files.tag_fetcher.error_handling import (
    handle_tag_errors,
)
from cg.services.deliver_files.tag_fetcher.models import DeliveryFileTags


class BamDeliveryTagsFetcher(FetchDeliveryFileTagsService):
    """Class to fetch tags for bam files to deliver."""

    @handle_tag_errors
    def fetch_tags(self, workflow: Workflow) -> DeliveryFileTags:
        """Fetch the tags for the bam files to deliver."""
        self._validate_workflow(workflow=workflow)
        return DeliveryFileTags(
            case_tags=None,
            sample_tags=[{AlignmentFileTag.BAM}],
        )

    @staticmethod
    def _validate_workflow(workflow: Workflow):
        """Validate the workflow."""
        if workflow != Workflow.RAW_DATA:
            raise ValueError(f"Workflow {workflow} is not supported for only BAM file delivery.")
        return workflow



""" Module for retrieving flow cells from backup."""

import logging
import subprocess
from pathlib import Path

from housekeeper.store.models import File

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.exc import ChecksumFailedError
from cg.services.pdc_service.pdc_service import PdcService
from cg.meta.encryption.encryption import SpringEncryptionAPI
from cg.models import CompressionData

LOG = logging.getLogger(__name__)


class SpringBackupAPI:
    """Class for handling PDC backup and retrieval of spring compressed fastq files."""

    def __init__(
        self,
        encryption_api: SpringEncryptionAPI,
        pdc_service: PdcService,
        hk_api: HousekeeperAPI,
        dry_run: bool = False,
    ):
        self.encryption_api: SpringEncryptionAPI = encryption_api
        self.hk_api: HousekeeperAPI = hk_api
        self.pdc: PdcService = pdc_service
        self.dry_run = dry_run

    def encrypt_and_archive_spring_file(self, spring_file_path: Path) -> None:
        """Encrypts and archives a spring file and its decryption key."""
        LOG.debug(f"*** START BACKUP PROCESS OF SPRING FILE {spring_file_path} ***")
        if self.is_compression_ongoing(spring_file_path):
            LOG.info(
                f"Spring (de)compression ongoing, skipping archiving for spring file {spring_file_path}",
            )
            return
        self.encryption_api.cleanup(spring_file_path)
        if self.is_spring_file_archived(spring_file_path):
            LOG.info(
                "Spring file already archived! Spring file will be removed from disk, continuing "
                "to next spring file "
            )
            self.remove_archived_spring_file(spring_file_path)
            return
        try:
            self.encryption_api.spring_symmetric_encryption(spring_file_path)
            self.encryption_api.key_asymmetric_encryption(spring_file_path)
            self.encryption_api.compare_spring_file_checksums(spring_file_path)
            self.pdc.archive_file_to_pdc(
                file_path=str(self.encryption_api.encrypted_spring_file_path(spring_file_path))
            )
            self.pdc.archive_file_to_pdc(
                file_path=str(self.encryption_api.encrypted_key_path(spring_file_path))
            )
            self.mark_file_as_archived(spring_file_path)
            self.encryption_api.cleanup(spring_file_path)
            self.remove_archived_spring_file(spring_file_path)
            LOG.debug("*** ARCHIVING PROCESS COMPLETED SUCCESSFULLY ***")
        except subprocess.CalledProcessError as error:
            LOG.error(f"Encryption failed: {error.stderr}")
            LOG.debug("*** COMMAND PROCESS FAILED! ***")
            self.encryption_api.cleanup(spring_file_path)
        except ChecksumFailedError as error:
            LOG.error(error)
            self.encryption_api.cleanup(spring_file_path)
            LOG.debug("*** CHECKSUM PROCESS FAILED! ***")

    def retrieve_and_decrypt_spring_file(self, spring_file_path: Path) -> None:
        """Retrieves and decrypts a spring file and its decryption key."""
        LOG.info(f"*** START RETRIEVAL PROCESS OF SPRING FILE {spring_file_path} ***")
        try:
            self.pdc.retrieve_file_from_pdc(
                file_path=str(self.encryption_api.encrypted_spring_file_path(spring_file_path)),
            )
            self.pdc.retrieve_file_from_pdc(
                file_path=str(self.encryption_api.encrypted_key_path(spring_file_path)),
            )
            self.encryption_api.key_asymmetric_decryption(spring_file_path)
            self.encryption_api.spring_symmetric_decryption(
                spring_file_path, output_file=spring_file_path
            )
        except subprocess.CalledProcessError as error:
            LOG.error(f"Decryption failed: {error.stderr}")
            LOG.debug("*** RETRIEVAL PROCESS FAILED! ***")
        self.encryption_api.cleanup(spring_file_path)
        LOG.debug("*** RETRIEVAL PROCESS COMPLETED SUCCESSFULLY ***")

    def mark_file_as_archived(self, spring_file_path: Path) -> None:
        """Set the field 'to_archive' of the file in Housekeeper to mark that it has been
        archived to PDC."""
        if self.dry_run:
            LOG.info(f"Dry run, no changes made to {spring_file_path}")
            return
        hk_spring_file: File = self.hk_api.files(path=str(spring_file_path)).first()
        if hk_spring_file:
            LOG.info(f"Setting {spring_file_path} to archived in Housekeeper")
            self.hk_api.set_to_archive(file=hk_spring_file, value=True)
        else:
            LOG.warning(f"Could not find {spring_file_path} on disk")

    def remove_archived_spring_file(self, spring_file_path: Path) -> None:
        """Removes all files related to spring PDC archiving."""
        if not self.dry_run:
            LOG.info(f"Removing spring file {spring_file_path} from disk")
            spring_file_path.unlink()

    def is_to_be_retrieved_and_decrypted(self, spring_file_path: Path) -> bool:
        """Determines if a spring file is archived on PDC and needs to be retrieved and decrypted."""
        spring_file: File = self.hk_api.files(path=str(spring_file_path)).first()
        if spring_file and not spring_file_path.exists():
            return spring_file.to_archive
        return False

    def is_spring_file_archived(self, spring_file_path: Path) -> bool:
        """Checks if a spring file is marked as archived in Housekeeper."""
        spring_file: File = self.hk_api.files(path=str(spring_file_path)).first()
        if spring_file:
            return spring_file.to_archive
        return False

    @staticmethod
    def is_compression_ongoing(spring_file_path: Path) -> bool:
        """Determines if (de)compression of the spring file ongoing."""
        return CompressionData(spring_file_path).pending_exists()



from cg.services.sample_run_metrics_service.sample_run_metrics_service import (
    SampleRunMetricsService,
)


def test_get_metrics_for_flow_cell(
    sample_run_metrics_service: SampleRunMetricsService,
    novaseq_x_flow_cell_id: str,
):
    # GIVEN an existing flow cell with sequencing metrics

    # WHEN fetching the metrics for the flow cell
    metrics = sample_run_metrics_service.get_metrics(novaseq_x_flow_cell_id)

    # THEN the metrics should be returned
    assert metrics



import os
from pathlib import Path
from cg.services.deliver_files.file_fetcher.models import SampleFile
from cg.services.deliver_files.file_formatter.models import FormattedFile


class FileManagingService:
    """
    Service to manage files.
    Handles operations that create or rename files and directories.
    """

    @staticmethod
    def create_directories(base_path: Path, directories: set[str]) -> None:
        """Create directories for given names under the base path."""
        for directory in directories:
            Path(base_path, directory).mkdir(exist_ok=True)

    @staticmethod
    def rename_file(src: Path, dst: Path) -> None:
        """Rename a file from src to dst."""
        os.rename(src, dst)


class SampleFileNameFormatter:
    """
    Class to format sample file names.
    """

    @staticmethod
    def get_sample_names(sample_files: list[SampleFile]) -> set[str]:
        """Extract sample names from the sample files."""
        return {sample_file.sample_name for sample_file in sample_files}

    @staticmethod
    def format_sample_file_names(sample_files: list[SampleFile]) -> list[FormattedFile]:
        """
        Returns formatted files with original and formatted file names:
        1. Adds a folder with sample name to the path of the sample files.
        2. Replaces sample id by sample name.
        """
        formatted_files = []
        for sample_file in sample_files:
            replaced_name = sample_file.file_path.name.replace(
                sample_file.sample_id, sample_file.sample_name
            )
            formatted_path = Path(
                sample_file.file_path.parent, sample_file.sample_name, replaced_name
            )
            formatted_files.append(
                FormattedFile(original_path=sample_file.file_path, formatted_path=formatted_path)
            )
        return formatted_files


class SampleFileFormatter:
    """
    Format the sample files to deliver.
    Used for all workflows except Microsalt and Mutant.
    """

    def __init__(
        self, file_manager: FileManagingService, file_name_formatter: SampleFileNameFormatter
    ):
        self.file_manager = file_manager
        self.file_name_formatter = file_name_formatter

    def format_files(
        self, moved_files: list[SampleFile], ticket_dir_path: Path
    ) -> list[FormattedFile]:
        """Format the sample files to deliver and return the formatted files."""
        sample_names: set[str] = self.file_name_formatter.get_sample_names(sample_files=moved_files)
        for sample_name in sample_names:
            self.file_manager.create_directories(
                base_path=ticket_dir_path, directories={sample_name}
            )
        formatted_files: list[FormattedFile] = self.file_name_formatter.format_sample_file_names(
            sample_files=moved_files
        )
        for formatted_file in formatted_files:
            self.file_manager.rename_file(
                src=formatted_file.original_path, dst=formatted_file.formatted_path
            )
        return formatted_files



from pathlib import Path

from cg.constants import FileExtensions
from cg.services.illumina.backup.utils import (
    get_latest_archived_encryption_key_path,
    get_latest_archived_sequencing_run_path,
)


def test_get_latest_encryption_key_path(dsmc_q_archive_output: list[str], flow_cell_name: str):
    """Tests returning an encryption key path from DSMC output."""
    # GIVEN an DSMC output and a flow cell id
    key_path: Path = get_latest_archived_encryption_key_path(dsmc_output=dsmc_q_archive_output)

    # THEN this method should return a path object
    assert isinstance(key_path, Path)

    # THEN the latest key file name should be returned
    expected: str = (
        f"161115_ST-E00214_0118_B{flow_cell_name}{FileExtensions.KEY}{FileExtensions.GPG}"
    )
    assert key_path.name == expected


def test_get_latest_archived_run_path(dsmc_q_archive_output: list[str], flow_cell_name: str):
    """Tests returning an Illumina run path from DSMC output."""
    # GIVEN an DSMC output and a flow cell id
    runs_path: Path = get_latest_archived_sequencing_run_path(dsmc_output=dsmc_q_archive_output)

    # THEN this method should return a path object
    assert isinstance(runs_path, Path)

    # THEN the latest Illumina run file name should be returned
    expected: str = (
        f"161115_ST-E00214_0118_B{flow_cell_name}{FileExtensions.TAR}{FileExtensions.GZIP}{FileExtensions.GPG}"
    )
    assert runs_path.name == expected



from pydantic import BaseModel


class SequencingMetrics(BaseModel):
    flow_cell_name: str
    flow_cell_lane_number: int
    sample_internal_id: str
    sample_total_reads_in_lane: int
    sample_base_percentage_passing_q30: float
